/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * For demonstration purpose only.
 * MicroEJ Corp. PROPRIETARY. Use is subject to license terms.
 */
package com.microej.demo.thermorest.model;

import java.util.concurrent.Executor;

import com.microej.demo.thermorest.util.JSONStringBuilder;

import ej.basictool.ArrayTools;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.util.concurrent.SingleThreadExecutor;

/**
 * A model describing a thermostat.
 */
public class ThermostatModel {

	private static final int DEFAULT_TEMPERATURE = 15;

	/**
	 * The target tag.
	 */
	public static final String TARGET = "target"; //$NON-NLS-1$

	/**
	 * The temperature tag.
	 */
	public static final String TEMPERATURE = "temperature"; //$NON-NLS-1$

	/**
	 * The unit tag.
	 */
	public static final String UNIT = "unit"; //$NON-NLS-1$

	/**
	 * HTML representation of a degree.
	 */
	public static final String DEG = "&deg;"; //$NON-NLS-1$

	/**
	 * Celcius unit.
	 */
	public static final String CELCIUS = DEG + 'C';

	/**
	 * Farenheight unit.
	 */
	public static final String FARENHEIGHT = DEG + 'F';

	private String unit = CELCIUS;
	private float temperature;
	private float target;
	private ThermostatListener[] listeners;
	private final Executor executor;

	/**
	 * Instantiates a {@link ThermostatModel}.
	 */
	public ThermostatModel() {
		this.temperature = DEFAULT_TEMPERATURE;
		this.target = DEFAULT_TEMPERATURE;
		this.listeners = new ThermostatListener[0];
		this.executor = ServiceLoaderFactory.getServiceLoader().getService(Executor.class, SingleThreadExecutor.class);
	}

	/**
	 * Gets the temperature.
	 *
	 * @return the temperature.
	 */
	public float getTemperature() {
		return this.temperature;
	}

	/**
	 * Sets the temperature and notify the listeners.
	 *
	 * @param temperature
	 *            the temperature to set.
	 */
	public void setTemperature(float temperature) {
		if (temperature != this.temperature) {
			this.temperature = temperature;
			notifyTemperatureChange(temperature);
		}
	}

	/**
	 * Gets the target temperature.
	 *
	 * @return the target temperature.
	 */
	public float getTargetTemperature() {
		return this.target;
	}

	/**
	 * Sets the target and notify the listeners.
	 *
	 * @param target
	 *            the target to set.
	 */
	public void setTargetTemperature(float target) {
		if (target != this.target) {
			this.target = target;
			notifyTargetChange(target);
		}
	}

	/**
	 * Append a thermostat listener.
	 *
	 * @param e
	 *            element to be appended to this list
	 */
	public void addListener(ThermostatListener e) {
		this.listeners = ArrayTools.add(this.listeners, e);
	}

	/**
	 * Removes a ThermostatListener.
	 *
	 * @param o
	 *            collection containing elements to be removed from this list
	 */
	public void removeListener(ThermostatListener o) {
		this.listeners = ArrayTools.remove(this.listeners, o);
	}

	/**
	 * Gets the listeners.
	 *
	 * @return the listeners.
	 */
	public ThermostatListener[] getListeners() {
		return this.listeners.clone();
	}

	/**
	 * Gets the unit.
	 *
	 * @return the unit.
	 */
	public String getUnit() {
		return this.unit;
	}

	/**
	 * Sets the unit.
	 *
	 * @param unit
	 *            the unit to set.
	 */
	public void setUnit(String unit) {
		this.unit = unit;
	}

	/**
	 * Gets a serialize version of the {@link ThermostatModel}.
	 *
	 * @return a String of the {@link ThermostatModel} in JSON format.
	 */
	public String serialize() {
		StringBuilder builder = new StringBuilder().append('{');
		JSONStringBuilder.append(builder, TEMPERATURE, this.temperature).append(',');
		JSONStringBuilder.append(builder, TARGET, this.target).append(',');
		JSONStringBuilder.append(builder, UNIT, this.unit);
		builder.append('}');
		return builder.toString();
	}

	@Override
	public String toString() {
		return serialize();
	}

	private void notifyTargetChange(final float target) {
		this.executor.execute(new Runnable() {

			@Override
			public void run() {
				for (ThermostatListener listener : ThermostatModel.this.listeners) {
					listener.onTargetChange(target);
				}
			}
		});
	}

	private void notifyTemperatureChange(final float temperature) {
		this.executor.execute(new Runnable() {

			@Override
			public void run() {
				for (ThermostatListener listener : ThermostatModel.this.listeners) {
					listener.onTemperatureChange(temperature);
				}
			}
		});
	}
}
