/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.kf.util.ecom;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import ej.ecom.Device;
import ej.ecom.HardwareDescriptor;
import ej.kf.Converter;
import ej.kf.Feature;
import ej.kf.Kernel;

/* A {@link Converter} for {@link HardwareDescriptor}.
 * This converter creates a clone of the original {@link HardwareDescriptor}.
 */
public class HardwareDescriptorConverter implements Converter<HardwareDescriptor>{
	@Override
	public HardwareDescriptor<?> convert(HardwareDescriptor hd, Feature targetOwner) throws IllegalAccessError {
		try {
			final String name = Kernel.clone(hd.getName(), targetOwner);
			final String [] propertyNames = hd.getPropertyNames();
			final int size = propertyNames.length;

			//			Create a map is the targetOwner context
			RunnableWithResult<Map<String, String>> runnableMap = new RunnableWithResult<Map<String, String>>() {
				@Override
				public Map<String, String> runWithResult() {
					return new HashMap<String, String>(size);
				}
			};

			// Fill that map with the properties of hd
			Kernel.runUnderContext(targetOwner, runnableMap);
			final Map<String, String> result = runnableMap.getResult();
			for (String property : propertyNames) {
				// clone the property and it value
				result.put(Kernel.clone(property, targetOwner), Kernel.clone(hd.getProperty(property), targetOwner));
			}

			// Create the result in the targetOwner context
			RunnableWithResult<PropertiesDescriptor> runnableHD = new RunnableWithResult<PropertiesDescriptor>() {
				@Override
				public PropertiesDescriptor runWithResult() {
					return new PropertiesDescriptor(name, result);
				}
			};
			Kernel.runUnderContext(targetOwner, runnableHD);
			return runnableHD.getResult();

		} catch (CloneNotSupportedException e) {
			// should not occur
			throw new AssertionError();
		}
	}

	@Override
	public Class<HardwareDescriptor> getType() {
		return HardwareDescriptor.class;
	}

	class PropertiesDescriptor<D extends Device> implements HardwareDescriptor<D> {

		private final Map<String, String> properties;
		private final String name;

		public PropertiesDescriptor(String name, Map<String, String> properties){
			this.name = name;
			this.properties = properties;
		}

		@Override
		public String getName() {
			return name;
		}

		@Override
		public String getProperty(String propertyName) {
			return properties.get(propertyName);
		}

		@Override
		public String[] getPropertyNames() {
			Set<String> keys = properties.keySet();
			return keys.toArray(new String[keys.size()]);
		}

		@Override
		public String[] getPropertyValues() {
			Collection<String> values = properties.values();
			return values.toArray(new String[values.size()]);
		}
	}

	public abstract class RunnableWithResult<R> implements Runnable {

		private R result;

		protected void setResult(R result) {
			Kernel.enter();
			this.result = result;
			// Kernel.exit();
		}

		/**
		 * Gets the result.
		 *
		 * @return the result.
		 */
		public R getResult() {
			return this.result;
		}

		/**
		 * By default sets the result returned by {@link #runWithResult()}.
		 */
		@Override
		public final void run(){
			setResult(runWithResult());
		}

		protected abstract R runWithResult();

	}
}
