/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.profiling;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.microej.profiling.internal.Strings;
import com.microej.profiling.internal.snapshot.PrintableSnapshot;
import com.microej.profiling.internal.snapshot.Snapshot;
import com.microej.profiling.profiler.ComparableProfiler;

/**
 *
 * A group of sampling used to gets the statistics values on a {@link SampleProfiler}.
 *
 */
public class Group {

	private final String name;
	private final Set<PrintableSnapshot> snapshots;

	/**
	 * Instantiates a {@link Group}.
	 *
	 * @param name
	 *            the group name.
	 */
	public Group(String name) {
		this.name = name;
		this.snapshots = new HashSet<>();
	}

	/**
	 * Gets the name.
	 *
	 * @return the name of the group.
	 */
	public String getName() {
		return this.name;
	}

	@Override
	public String toString() {
		Map<ComparableProfiler, Statitics> samplers = new HashMap<>();
		for (PrintableSnapshot printableSnapshot : this.snapshots) {
			Snapshot snapshot = printableSnapshot.getStart();
			ComparableProfiler profiler = snapshot.getProfiler();
			Statitics value = samplers.get(profiler);
			if (value == null) {
				value = new Statitics();
				samplers.put(profiler, value);
			}
			value.add(profiler.compare(snapshot.getSample(), printableSnapshot.getEnd().getSample()));
		}
		StringBuilder out = new StringBuilder(Strings.TAB + getName() + Strings.COLLON + Strings.LINE_BREAK);
		for (Entry<ComparableProfiler, Statitics> entry : samplers.entrySet()) {
			ComparableProfiler sampler = entry.getKey();
			Statitics value = entry.getValue();

			long min = value.getMin();
			long max = value.getMax();
			if (sampler.isBiggerPrinted(min, max)) {
				long tmp = min;
				min = max;
				max = tmp;
			}
			out.append(Strings.TAB);
			out.append(Strings.MINUS);
			out.append(sampler.getName());
			out.append(Strings.COLLON);
			out.append(Strings.TAB);
			out.append(Strings.COUNT);
			out.append(value.getCount());
			out.append(Strings.COMMA);
			out.append(Strings.MIN);
			out.append(sampler.printValue(min));
			out.append(Strings.COMMA);
			out.append(Strings.MAX);
			out.append(sampler.printValue(max));
			out.append(Strings.COMMA);
			out.append(Strings.AVG);
			out.append(sampler.printValue(value.getAverage()));
			out.append(Strings.LINE_BREAK);
		}
		return out.toString();
	}

	/**
	 * Adds a snapshot.
	 *
	 * @param printableSnapshot
	 *            the snapshot to add.
	 */
	public void add(PrintableSnapshot printableSnapshot) {
		synchronized (this.snapshots) {
			this.snapshots.add(printableSnapshot);
		}
	}

	/**
	 * Clears the snapshots.
	 */
	public void clear() {
		synchronized (this.snapshots) {
			this.snapshots.clear();
		}
	}

	private static class Statitics {
		private int count;
		private long sum;
		private long max;
		private long min;

		public void add(long value) {
			if (this.count == 0) {
				this.max = value;
				this.min = value;
			}
			this.max = Math.max(value, this.max);
			this.min = Math.min(value, this.min);
			this.count++;
			this.sum += value;
		}

		public long getAverage() {
			return this.sum / this.count;
		}

		public int getCount() {
			return this.count;
		}

		public long getMin() {
			return this.min;
		}

		public long getMax() {
			return this.max;
		}
	}
}
