/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.profiling.profiler;

import java.util.logging.Logger;

import com.microej.profiling.internal.Strings;

/**
 * Profile the memory usage.
 */
public abstract class MemoryProfiler implements ComparableProfiler, InstantProfiler {

	private static final String UNIT = Strings.BYTE;

	/**
	 * The type of memory.
	 */
	public enum MemoryType {
		/**
		 * Display the free memory.
		 */
		FREE,
		/**
		 * Display the used memory.
		 */
		USED
	}

	private static MemoryType memoryType = MemoryType.USED;

	/**
	 * Sets the {@link MemoryType} to use.
	 *
	 * @param memoryType
	 *            the {@link MemoryType} to use.
	 */
	public static void setMemoryType(MemoryType memoryType) {
		MemoryProfiler.memoryType = memoryType;
	}

	/**
	 * Prints the memory to logger.
	 *
	 * @param logger
	 *            the logger to use.
	 * @param total
	 *            the total memory.
	 * @param freeMemory
	 *            the free memory.
	 * @param printUsedMemory
	 *            <code>true</code> if the memory to print is the used memory.
	 * @param printTotal
	 *            <code>true</code> to print the total memory.
	 */
	protected static void printMemory(Logger logger, long total, long freeMemory, boolean printUsedMemory,
			boolean printTotal) {
		logger.info(printMemory(total, getMemory(total, freeMemory, printUsedMemory), printTotal));
	}

	/**
	 * Gets the memory to a string.
	 *
	 * @param total
	 *            the total memory available.
	 * @param memory
	 *            the memory to print.
	 * @param printTotal
	 *            <code>true</code> to print the total memory.
	 * @return A string with the memory.
	 */
	protected static String printMemory(long total, long memory, boolean printTotal) {
		StringBuffer buffer = new StringBuffer();
		buffer.append(memory).append(UNIT);
		if (printTotal) {
			buffer.append(Strings.SPLIT).append(total);
		}
		return buffer.toString();
	}

	/**
	 * Computes the used or free memory.
	 *
	 * @param total
	 *            the total memory available.
	 * @param free
	 *            the free memory.
	 * @param usedMemory
	 *            <code>true</code> if the memory to print is the used memory.
	 * @return the memory to print.
	 */
	protected static long getMemory(long total, long free, boolean usedMemory) {
		return usedMemory ? total - free : free;
	}

	@Override
	public String printValue(long value) {
		if (memoryType == MemoryType.USED) {
			long totalMemory = getTotal();
			value = totalMemory - value;
		}
		return value + UNIT;
	}

	@Override
	public String getName() {
		String type = (memoryType == MemoryType.USED) ? Strings.USED : Strings.FREE;
		return type + getType();
	}

	@Override
	public long compare(long start, long end) {
		long diff;
		if (memoryType == MemoryType.USED) {
			long totalMemory = getTotal();
			diff = end - start + totalMemory;
		} else {
			diff = end - start;
		}
		return diff;
	}

	@Override
	public boolean isBiggerPrinted(long v1, long v2) {
		boolean bigger = v1 > v2;
		if (memoryType == MemoryType.USED) {
			bigger = !bigger;
		}
		return bigger;
	}

	/**
	 * Gets the total memory available.
	 *
	 * @return the total memory available.
	 */
	protected abstract long getTotal();

	/**
	 * Gets the type of memory.
	 *
	 * @return the type of memory.
	 */
	protected abstract String getType();
}
