/*
 * Java 1.7
 *
 * Copyright 2017-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.security.command;

import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import com.microej.security.util.DerFormatException;
import com.microej.security.util.DerInputStreamHelper;
import com.microej.security.util.DerValueInputStream;

/**
 * This implementation of {@link CommandUnwrapper} DOES NOT verify the signature of the SignedCommand. It simply skips
 * algId, certPath to reach the command content and closes the stream before reading the signature.
 */
@SuppressWarnings("nls")
public class NoVerifySignedCommandUnwrapper implements CommandUnwrapper {
	private InputStream is;
	private static final int VERSION = 1;
	private static final byte COMMAND_TAG_MASK = DerInputStreamHelper.MASK_CONTEXT
			| DerInputStreamHelper.MASK_CONSTRUCTED;

	@Override
	public Command unwrapCommand(InputStream wrappedCommandStream, final IProgressMonitor monitor)
			throws IOException, InvalidCommandFormatException {
		// The stream may be preceded by an explicit tag (0xA0 or 0xA1)
		this.is = wrappedCommandStream;

		final SubMonitor progress = SubMonitor.convert(monitor, 100);
		progress.subTask("Checking headers");

		int tag = wrappedCommandStream.read();
		try {
			// Consumes the first bytes (length of the tag A0 | A1)
			/* int length = */ DerInputStreamHelper.readDERLength(wrappedCommandStream);
			if (tag != DerInputStreamHelper.tag_Sequence) {
				// Ensure that the stream is wrapped in a sequence
				/* int totalLength = */ DerInputStreamHelper.expectTag(DerInputStreamHelper.tag_Sequence,
						wrappedCommandStream);
			}

			// Read and validate Version:
			checkVersion();
			progress.worked(5);

			skipAlgId();
			progress.worked(5);

			progress.subTask("Skipping certpath");
			skipCertPath();
			progress.worked(15);

			// Return the command
			DerValueInputStream commandInputStream = new DerValueInputStream(this.is) {
				@Override
				protected void onDataRead(byte[] data, int off, int len) throws InvalidSignatureException {
					progress.worked(len);
				}

				@Override
				protected void onDataRead(byte b) throws InvalidSignatureException {
					progress.worked(1);
				}

				@Override
				protected void onStreamEnd() throws IOException {
					progress.subTask("Skipping signature");

					// Consumes mindlessly the signature bits
					int read;
					do {
						read = read();
					} while (read != -1);
					if (monitor != null) {
						monitor.done();
					}
				}
			};
			int commandId = commandInputStream.getTag() & (~COMMAND_TAG_MASK);
			return new Command(commandInputStream, commandId);
		} catch (DerFormatException e) {
			throw new InvalidCommandFormatException(e);
		}
	}

	private void skipCertPath() throws IOException, DerFormatException {
		int length = DerInputStreamHelper.expectTag(DerInputStreamHelper.tag_Sequence, this.is);
		this.is.skip(length);
	}

	private void skipAlgId() throws IOException, DerFormatException {
		int length = DerInputStreamHelper.expectTag(DerInputStreamHelper.tag_Sequence, this.is);
		this.is.skip(length);
	}

	private void checkVersion() throws IOException, DerFormatException {
		int version = DerInputStreamHelper.readInteger(this.is);
		if (NoVerifySignedCommandUnwrapper.VERSION != version) {
			throw new IOException("Stream version not supported. Support v" + NoVerifySignedCommandUnwrapper.VERSION
					+ " but stream version is v" + version + ".");
		}
	}
}
