/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.basedriver.impl;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import ej.basedriver.Controller;
import ej.basedriver.ControllerListener;
import ej.basedriver.Node;
import ej.ecom.Device;
import ej.ecom.HardwareDescriptor;


/**
 * This abstract class gives a structure and mechanisms for developers who wish to create a {@link Controller} for a protocol.
 * 
 * @param <ConcreteNode>
 *            the type of the nodes in the network managed by this controller
 */
public abstract class AbstractController<ConcreteNode extends Node> implements Controller {

	/**
	 * The listener to notify when devices are included in and excluded from the network managed by this controller, when an error occurs when
	 * communicating with a device, when there is an event on device.
	 */
	protected ControllerListener listener;
	/**
	 * List of the nodes known by this controller.
	 */
	protected Map<Integer, ConcreteNode> nodes;

	private boolean started;

	/**
	 * The controller
	 */
	private int state;

	protected PropertiesDescriptor<Controller> descriptor;

	/**
	 * Create a new controller. A controller is an {@link Device} so its has a UID. This UID may be used to identify the managed subnetwork in
	 * the global network.
	 */
	protected AbstractController() {
		this("unknown");
	}

	/**
	 * Create a new controller with the given protocol name which will be its {@link HardwareDescriptor#getName()}.
	 * @param name the protocol name
	 */
	protected AbstractController(String name) {
		this(name, null);
	}

	/**
	 * Create a new controller from a source {@link Device}.
	 * Descriptor of the source device is copied to the controller under its <code>{@link HardwareDescriptor#getName()}/</code> group.
	 * @param name the protocol name
	 * @param source source device on which is created this {@link Controller}, null if controller is not created from a {@link Device} source
	 */
	protected AbstractController(String name, Device source) {
		this.nodes = new HashMap<>();
		this.descriptor = new PropertiesDescriptor<Controller>(name);
		if(source != null){
			HardwareDescriptor<Device> sourceDesc = source.getDescriptor();
			String sourceName = sourceDesc.getName();
			for(String key : sourceDesc.getPropertyNames()){
				String value = sourceDesc.getProperty(key);
				if(value == null){
					// device is dead
					return;
				}
				String newKey = new StringBuilder(sourceName).append('/').append(key).toString();
				descriptor.setProperty(newKey, value);
			}
		}
	}

	@Override
	public String getName() {
		return descriptor.getName();
	}

	protected void setName (String name) {
		descriptor.setName(name);
	}

	@SuppressWarnings("unchecked")
	@Override
	public HardwareDescriptor<Controller> getDescriptor() {
		return descriptor;
	}

	/**
	 * When a node is paired, it must be added to the controller so that the controller knows all paired nodes.
	 * @param uid
	 * 			node id.
	 * @param node
	 *            a new node
	 */
	public void addNode(int uid, ConcreteNode node) {
		nodes.put(uid, node);
	}

	/**
	 * Get a particular node paired to this controller.
	 * 
	 * @param uid
	 *            UID on the node inside the UNET network of this controller
	 * @return the node with this UID or null if no such node was found
	 */
	public ConcreteNode getNode(int uid) {
		return nodes.get(uid);
	}

	/**
	 * Remove a node.
	 * 
	 * @param uid
	 *            the UID of the node to exclude.
	 */
	public void removeNode(int uid) {
		ConcreteNode node = getNode(uid);
		if (node != null) {
			// Remove the node from our list
			unregisterNode(node);
			nodes.remove(uid);
		}
	}

	private void unregisterNode (ConcreteNode node) {
		// Unregister of the node's children
		Device[] childs = node.getChildren();
		if (childs != null) {
			for (Device device : childs) {
				try {
					listener.onUnregister(device);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		// Unregister the node itself
		try {
			listener.onUnregister(node);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Remove all nodes.
	 */
	private void removeAllNodes () {
		Iterator<Integer> it = nodes.keySet().iterator();
		while(it.hasNext()) {
			unregisterNode(nodes.get(it.next()));
			it.remove();
		}
	}

	@Override
	public ControllerListener getListener() {
		return listener;
	}

	@Override
	public Device getParent() {
		return null; // a controller doesn't have a parent
	}

	@Override
	public Device[] getChildren() {
		return nodes.values().toArray(new Device[nodes.size()]);
	}

	@Override
	public boolean start(ControllerListener listener) {
		if (listener == null) {
			throw new NullPointerException("The listener cannot be null");
		}

		if (started) {
			throw new IllegalStateException();
		}

		state = RUNNING;
		this.listener = listener;
		boolean ok = startController();
		if (ok) {
			started = true;
			try {
				defaultControllerInit();
				return true;
			} catch (Throwable e) {
			}
		}
		return false;
	}

	/**
	 * Default Controller initialization sequence: the {@link Controller} is registered before calling {@link #initialize()}.
	 * @throws IOException
	 *             if an error occurs on streams
	 */
	protected void defaultControllerInit() throws IOException{
		registerController();
		initialize();
	}

	/**
	 * Triggers the registration of the {@link Controller} to the {@link #listener}.
	 */
	public void registerController() {
		listener.onRegister(Controller.class, this);
	}

	@Override
	public synchronized void stop() {
		if (started) {
			stopController();
			removeAllNodes();
			try {
				listener.onUnregister(this);
			} catch (Exception e) {
				e.printStackTrace();
			}
			started = false;
		} else {
			throw new IllegalStateException();
		}
	}

	/**
	 * Returns true if the controller is started, false otherwise.
	 * @return true if the controller is started, false otherwise.
	 */
	@Override
	public synchronized boolean isStarted () {
		return started;
	}

	/**
	 * Do specific actions to start the concrete controller.
	 * 
	 * @return true if the controller has been successfully started, false otherwise.
	 */
	protected abstract boolean startController();

	/**
	 * Do specific actions to stop the concrete controller.
	 */
	protected abstract void stopController();

	/**
	 * Request all known nodes and profiles.
	 * 
	 * @throws IOException
	 *             if an error occurs on streams
	 */
	protected abstract void initialize() throws IOException;

	@Override
	public void inclusionStart() {
		if (state != RUNNING) {
			throw new IllegalArgumentException("The controller is not in RUNNING state");
		}

		inclusionStartController();
	}

	/**
	 * Start the inclusion phase.
	 */
	protected abstract void inclusionStartController();

	@Override
	public void inclusionStop() {
		if (state != INCLUSION) {
			throw new IllegalArgumentException("The controller is not in INCLUSION state");
		}

		inclusionStopController();
	}

	/**
	 * Stop the inclusion phase.
	 */
	protected abstract void inclusionStopController();


	@Override
	public void exclusionStart() {
		if (state != RUNNING) {
			throw new IllegalArgumentException("The controller is not in RUNNING state");
		}

		exclusionStartController();
	}

	/**
	 * Start the exclusion phase.
	 */
	protected abstract void exclusionStartController();

	@Override
	public void exclusionStop() {
		if (state != EXCLUSION) {
			throw new IllegalArgumentException("The controller is not in EXCLUSION state");
		}

		exclusionStopController();
	}

	/**
	 * Stop the exclusion phase.
	 */
	protected abstract void exclusionStopController();

	@Override
	public int getState() {
		return state;
	}

	/**
	 * Change the current state of the controller.
	 * Notify the controller listener if the newState is different from the current
	 * @param newState
	 * 					The new state of the controller
	 */
	public void setState (int newState) {
		if (state != newState) {
			state = newState;
			listener.onEvent(Controller.class, this);
		}
	}
}
