/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.basedriver.zwave;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import ej.basedriver.impl.AbstractController;
import ej.basedriver.zwave.frame.GetInitialData;
import ej.basedriver.zwave.frame.MemoryGetID;
import ej.basedriver.zwave.frame.RequestFrame;
import ej.basedriver.zwave.frame.RequestFrameListener;
import ej.basedriver.zwave.util.ZwaveLogger;
import ej.basedriver.zwave.util.ZwaveLoggerAdapter;
import ej.ecom.Device;
import ej.ecom.io.StreamConnection;

public class ZwaveController extends AbstractController<ZwaveNode> implements RequestFrameListener {
	public static final boolean DEBUG = false;

	private static final String ZWAVE_DESCRIPTOR_NAME = "zwave";

	/**
	 * The connection to communicate this the hardware controller.
	 */
	private final StreamConnection connection;

	private final ZwaveReceiveThread receiveThread;
	private final ZwaveSendThread sendThread;
	private ZwaveSession session;
	private byte id;
	private int homeId;
	private final ZwaveLogger logger;

	public ZwaveController(StreamConnection streamConnection) {
		this(streamConnection, new ZwaveLoggerAdapter());
	}

	public ZwaveController(StreamConnection streamConnection, ZwaveLogger logger) {
		this(null, streamConnection, logger);
	}

	public ZwaveController(Device source, StreamConnection streamConnection, ZwaveLogger logger) {
		super(ZWAVE_DESCRIPTOR_NAME, source);
		this.connection = streamConnection;
		this.logger = logger;
		sendThread = new ZwaveSendThread();
		receiveThread = new ZwaveReceiveThread(sendThread, logger);
	}

	public ZwaveController(Device source, StreamConnection streamConnection){
		this(source, streamConnection, new ZwaveLoggerAdapter());
	}

	protected ZwaveSession createSession () {
		return new ZwaveSession(this);
	}

	public ZwaveLogger getLogger () {
		return logger;
	}

	@Override
	protected void inclusionStartController() {
		sendThread.inclusionStart();
	}

	@Override
	protected void inclusionStopController() {
		sendThread.inclusionStop();
	}

	@Override
	public int getExclusionMode() {
		return MANUAL;
	}

	@Override
	protected void exclusionStartController() {
		sendThread.exclusionStart();
	}

	@Override
	protected void exclusionStopController() {
		sendThread.exclusionStop();
	}

	@Override
	public boolean startController() {
		InputStream inputStream;
		OutputStream outputStream;
		try {
			inputStream = this.connection.openInputStream();
			outputStream = this.connection.openOutputStream();
		} catch (IOException e) {
			return false;
		}

		receiveThread.setInputStream(inputStream);
		sendThread.setOutputStream(outputStream);

		session = createSession();
		receiveThread.setZwaveSession(session);
		sendThread.setZwaveSession(session);

		sendThread.start();
		receiveThread.start();
		return true;
	}

	@Override
	public void stopController() {
		receiveThread.halt();
		sendThread.halt();
		try { connection.close(); }
		catch (Exception e1) { e1.printStackTrace(); }
	}

	@Override
	protected void defaultControllerInit() throws IOException{
		initialize();
	}

	/**
	 * @param id
	 */
	public void setId(byte id) {
		this.id = id;
		descriptor.setProperty("Node-ID", Integer.toString(id));
	}

	/**
	 * @param homeId
	 */
	public void setHomeId(int homeId) {
		this.homeId = homeId;
		descriptor.setProperty("Home-ID", Integer.toHexString(homeId));
		setName(ZWAVE_DESCRIPTOR_NAME + '-' + Integer.toHexString(homeId));
	}

	/**
	 * @return the homeId
	 */
	public int getHomeId() {
		return homeId;
	}

	/**
	 * @return the controllerID
	 */
	public byte getId() {
		return id;
	}

	@Override
	protected void initialize () throws IOException {
		//		softReset();
		this.sendFrame(new MemoryGetID(this));
		this.sendFrame(new GetInitialData(this));
	}

	@Override
	public void notifyComplete() {
	}

	@Override
	public void notifyError(RequestFrameError error) {
	}

	/**
	 * @param frame
	 * @throws IOException
	 */
	public void sendFrame(RequestFrame frame) throws IOException {
		try {
			boolean pushed = sendThread.enqueue(frame);

			if (! pushed) {
				throw new IOException();
			}
		} catch (InterruptedException e) {
			throw new IOException(e);
		}
	}

	public void setAwakeNode(ZwaveNode node) {
		node.setAwake(true);
		sendThread.addAwakeNode(node);
	}
}