/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.basedriver.zwave.util;

/**
 * Blocking FIFO queue with fixed size (blocking push when queue is full).
 * @param <T> the type of elements in this queue
 * We assume that we have only one reader
 */
public class Queue<T> {
	//Queue
	private static final String QUEUE_SIZE = "queue.size";
	private static final String QUEUE_THROW_FULL = "queue.throw.full";
	private static final int DEFAULT_QUEUE_SIZE = 16;

	private final Object readMonitor;
	private final Object writeMonitor;
	private final int queueLength;

	private Object[] queue;

	private int ptrBegin;
	private int ptrEnd;
	private int count;
	private boolean wakeupReader;

	private final boolean blockOnFull;

	/**
	 * Constructs an empty queue with the specified initial capacity.
	 *
	 * @throws IllegalArgumentException if the specified initial capacity
	 *         is negative
	 */
	public Queue() {
		this(Integer.getInteger(QUEUE_SIZE, DEFAULT_QUEUE_SIZE), !Boolean.getBoolean(QUEUE_THROW_FULL));
	}

	/**
	 * Constructs an empty queue with the specified initial capacity.
	 *
	 * @param initialCapacity  the initial capacity of the list
	 * @param blockOnFull If it is true, block the {@link Queue#push(Object)} is the queue is full.
	 * @throws IllegalArgumentException if the specified initial capacity
	 *         is negative
	 */
	public Queue(int initialCapacity, boolean blockOnFull) {
		this.blockOnFull = blockOnFull;
		this.readMonitor = new Object();
		this.writeMonitor = new Object();
		if (initialCapacity <= 0) {
			throw new IllegalArgumentException();
		}
		// +1: one index in the queue is always empty
		this.queue = new Object[queueLength = (initialCapacity + 1)];
		this.wakeupReader = false;
	}

	/**
	 * Returns the numbers of elements in this queue.
	 * @return the numbers of elements in this queue
	 */
	public int getSize(){
		return queueLength - 1; // see constructor comment
	}

	@SuppressWarnings("unchecked")
	private T elementData (int index) {
		return (T) queue[ptrBegin];
	}

	/**
	 * Takes an object in the queue and returns it. If the queue is empty, blocks
	 * until an object is pushed. May return <code>null</code> if the queue is
	 * stopped (see {@link #stop()}).
	 *
	 * @return an object in the queue
	 * @throws InterruptedException
	 */
	// This method reads ptrEnd and modifies the prtBegin
	public T pop() throws InterruptedException {
		T event;
		synchronized (readMonitor) {
			while (queue != null && ptrBegin == ptrEnd && ! wakeupReader) {
				readMonitor.wait();
			}
			if (queue == null) {
				// queue has been stopped
				return null;
			}
			if (wakeupReader) {
				wakeupReader = false;
				return null;
			}

			event = elementData(ptrBegin);
			queue[ptrBegin] = null;
			ptrBegin = (ptrBegin + 1) % queueLength;
			count --;
		}
		// notify a blocked writer that a queue element is free.
		synchronized (writeMonitor) {
			writeMonitor.notify();
		}
		return event;
	}

	/**
	 * Non-blocking pop. Takes an object in the queue and returns it. If the
	 * queue is empty, returns <code>null</code>.
	 *
	 * @return an object in the queue or <code>null</code> if the queue is
	 *         empty.
	 */
	public T tryPop() {
		T event;
		synchronized (readMonitor) {
			// if queue stopped or empty
			if (queue == null || ptrBegin == ptrEnd) {
				return null;
			}
			event = elementData(ptrBegin);
			queue[ptrBegin] = null;
			ptrBegin = (ptrBegin + 1) % queueLength;
			count --;
		}
		// notify a blocked writer that a queue element is free.
		synchronized (writeMonitor) {
			writeMonitor.notify();
		}
		return event;
	}

	/**
	 * Adds an object in the queue. If the queue is full, blocks until an object
	 * is popped, if {@link #blockOnFull}
	 *
	 * @param value
	 *            the object to add in the queue.
	 * @throws InterruptedException
	 * @return false if queue is full or has been stopped
	 */
	// This method reads ptrBegin and modifies the prtEnd
	public boolean push(T value) throws InterruptedException {
		synchronized (writeMonitor) {
			// wait while the queue is full
			int nextEnd = 0;
			// nextEnd must be recomputed each time since another writer thread
			// may change ptrEnd while the current writer thread is sleeping
			while (queue != null
					&& (nextEnd = (ptrEnd + 1) % queueLength) == ptrBegin) {
				if(blockOnFull) {
					writeMonitor.wait();
				} else {
					return false;
				}
			}
			if (queue == null) {
				// queue has been stopped
				return false;
			}
			queue[ptrEnd] = value;
			ptrEnd = nextEnd;
			count ++;
		}
		synchronized (readMonitor) {
			readMonitor.notify();
		}
		return true;
	}

	/**
	 * Stop the queue.
	 * <p>
	 * Reader threads (threads blocked in the {@link #pop()} method) are
	 * notified and the pop methods return <code>null</code>. Writer threads
	 * (threads blocked in the {@link #push(Object)} method) are notified and
	 * the added object is ignored.
	 * </p>
	 * <p>
	 * Once the queue has been stopped, subsequent calls to {@link #pop()}
	 * returns <code>null</code>, and subsequent calls to {@link #push(Object)}
	 * are ignored.
	 * </p>
	 */
	// this method is the only one to acquire both reader and writer monitor
	// simultaneously.
	public void stop() {
		synchronized (writeMonitor) {
			synchronized (readMonitor) {
				queue = null;
				readMonitor.notifyAll();
			}
			writeMonitor.notifyAll();
		}
	}

	/**
	 * Returns <tt>true</tt> if this queue is stopped.
	 *
	 * @return <tt>true</tt> if this queue is stopped
	 */
	public boolean isStopped() {
		return queue == null;
	}

	/**
	 * Returns the size of this queue.
	 * @return the size of this queue.
	 */
	public int size () {
		return count;
	}

	/*
	 * Wakeup the reader if it's waiting.
	 */
	public void notifyReader () {
		synchronized (readMonitor) {
			wakeupReader = true;
			readMonitor.notify();
		}
	}
}
