/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.data.cbor;

import java.io.InputStream;

import ej.cbor.CborConstants;
import ej.cbor.CborDecoder;
import ej.cbor.CborType;
import ej.data.AbstractDataReader;
import ej.data.DataReader;

/**
 * Implementation of a {@link DataReader} for a CBOR stream.
 */
public class DataReaderCBOR extends AbstractDataReader{

	CborDecoder decoder;

	/**
	 * Creates a new {@link DataReader} on the given {@link InputStream}.
	 * @param is the {@link InputStream} from which the CBOR structured data is read.
	 */
	public DataReaderCBOR(InputStream is) {
		decoder = new CborDecoder(is);
	}

	@Override
	protected boolean hasMoreElements0() throws Throwable {
		CborType type = decoder.peekType();
		return type != null && !isBreak(type);
	}

	private boolean isBreak(CborType type) {
		return type.getMajorType() == CborConstants.TYPE_FLOAT_SIMPLE && type.getAdditionalInfo() == CborConstants.BREAK;
	}

	@Override
	protected boolean readBooleanValue0() throws Throwable {
		return decoder.readBoolean();
	}

	@Override
	protected int readIntValue0() throws Throwable {
		return (int)decoder.readInt();
	}

	@Override
	protected long readLongValue0() throws Throwable {
		return decoder.readInt();
	}

	@Override
	protected double readDoubleValue0() throws Throwable {
		return decoder.readDouble();
	}

	@Override
	protected String readStringValue0() throws Throwable {
		return decoder.readTextString();
	}

	@Override
	protected String readStringKey0() throws Throwable {
		return decoder.readTextString();
	}

	@Override
	protected int readIntKey0() throws Throwable {
		return (int)decoder.readInt();
	}

	@Override
	protected void readArrayValueStart0() throws Throwable {
		decoder.readArrayLength();
	}

	@Override
	protected void readArrayValueEnd0() throws Throwable {
		decoder.readBreak();
	}

	@Override
	protected void readMapValueStart0() throws Throwable {
		decoder.readMapLength();
	}

	@Override
	protected void readMapValueEnd0() throws Throwable {
		decoder.readBreak();
	}

	@Override
	public void skipValue0() throws Throwable {
		CborType type = decoder.peekType();
		int majorType = type.getMajorType();
		if(majorType == CborConstants.TYPE_ARRAY){
			readArrayValueStart0();
			while(hasMoreElements0()){
				skipValue0();
			}
			readArrayValueEnd0();
		}
		else if(majorType == CborConstants.TYPE_MAP){
			readMapValueStart0();
			while(hasMoreElements0()){
				readStringKey0();
				skipValue0();
			}
			readMapValueEnd0();
		}
		else if(majorType == CborConstants.TYPE_UNSIGNED_INTEGER || majorType == CborConstants.TYPE_NEGATIVE_INTEGER){
			decoder.readInt();
		}
		else if(majorType == CborConstants.TYPE_TEXT_STRING){
			decoder.readTextString();
		}
		else if(majorType == CborConstants.TYPE_FLOAT_SIMPLE){
			int info = type.getAdditionalInfo();
			if(info == CborConstants.DOUBLE_PRECISION_FLOAT){
				decoder.readDouble();
			}
			else if(info == CborConstants.TRUE || info == CborConstants.FALSE){
				decoder.readBoolean();
			}
			else{
				throw new AssertionError();
			}
		}
		else{
			throw new AssertionError();
		}
	}
}
