/*
 * Java
 *
 * Copyright 2017-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.data.pubsub.paho;

import java.io.ByteArrayInputStream;
import java.util.Set;

import org.eclipse.paho.client.mqttv3.IMqttAsyncClient;
import org.eclipse.paho.client.mqttv3.IMqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttCallback;
import org.eclipse.paho.client.mqttv3.MqttMessage;

import ej.data.DataReader;

/**
 * Callback registered in the {@link IMqttAsyncClient} used by the {@link PahoMqttAgent}.<p>
 * It is in charge of dispatching received message to the different {@link PahoMqttContext}.
 */
public class PahoMqttAgentCallback implements MqttCallback {

	private final PahoMqttAgent mqttAgent;

	/**
	 * @param mqttAgent the {@link PahoMqttAgent} of this callback
	 */
	public PahoMqttAgentCallback(PahoMqttAgent mqttAgent) {
		super();
		this.mqttAgent = mqttAgent;
	}

	@Override
	public void connectionLost(Throwable cause) {
		// Nothing to do. May be manage later.
	}

	@Override
	public void messageArrived(String topic, MqttMessage message) throws Exception {
		MqttAgentConfiguration agentConfiguration = mqttAgent.agentConfiguration;
		byte[] messagePayload = message.getPayload();
		Set<PahoMqttContext> contexts = mqttAgent.getMqttContextSubscribedTo(topic);
		for(PahoMqttContext context : contexts){
			// Create an InputStream that reads the data from the MQTT message.
			ByteArrayInputStream messageInputStream = new ByteArrayInputStream(messagePayload);
			DataReader reader = agentConfiguration.newDataReader(messageInputStream, topic);
			try{
				context.listener.messageReceived(topic, reader);
			}
			catch(Throwable t){
				// Error while calling the listener: notify the agent configuration.
				agentConfiguration.errorOnMessageReceivedCallback(t, context, topic, message);
			}
		}
	}

	@Override
	public void deliveryComplete(IMqttDeliveryToken token) {
		// Nothing to do.
	}

}
