/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.background;

import ej.annotation.Nullable;
import ej.microui.display.GraphicsContext;
import ej.style.border.SimpleRoundedBorder;
import ej.style.container.Rectangle;

/**
 * Draws a plain color on the background with a round corner.
 * <p>
 * Should be used in association with a {@link SimpleRoundedBorder} to:
 * <ul>
 * <li>anti-alias its edges by using the same color,</li>
 * <li>draw a colored border that fits its shape.</li>
 * </ul>
 * For a better rendering, the border radius must be a bit smaller than the background one.
 */
public class SimpleRoundedPlainBackground implements Background {

	private int cornerRadius;

	/**
	 * Creates a simple rounded plain background.
	 */
	public SimpleRoundedPlainBackground() {
	}

	/**
	 * Creates a simple rounded plain background specifying its corner radius.
	 *
	 * @param cornerRadius
	 *            the corner radius to set.
	 */
	public SimpleRoundedPlainBackground(int cornerRadius) {
		this.cornerRadius = cornerRadius;
	}

	/**
	 * Gets the corner radius.
	 *
	 * @return the corner radius.
	 */
	public int getCornerRadius() {
		return this.cornerRadius;
	}

	/**
	 * Sets the corner radius.
	 *
	 * @param cornerRadius
	 *            the corner radius to set.
	 */
	public void setCornerRadius(int cornerRadius) {
		this.cornerRadius = cornerRadius;
	}

	@Override
	public boolean isTransparent() {
		// Is transparent because it is usually associated with an anti-aliased border
		// and because its corners are not drawn (if the element is moved, its corners are not repainted).
		return true;
	}

	@Override
	public void apply(GraphicsContext g, Rectangle rectangle, int color) {
		int width = rectangle.getWidth();
		int height = rectangle.getHeight();

		// Since draw ellipse arc is not available, both radius have the same size.
		int cornerRadius = Math.min(this.cornerRadius, (Math.min(width, height) - 1) >> 1);

		g.setColor(color);
		g.fillRoundRect(1, 1, width - 2, height - 2, cornerRadius << 1, cornerRadius << 1);

		// Set background color to optimize future anti-aliased drawings.
		g.setBackgroundColor(color);
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof SimpleRoundedPlainBackground) {
			SimpleRoundedPlainBackground background = (SimpleRoundedPlainBackground) obj;
			return this.cornerRadius == background.cornerRadius;
		}
		return false;
	}

	@Override
	public int hashCode() {
		// Since the fields are not final, they can't be used to compute a constant hashcode.
		return 0;
	}

}
