/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.font;

import ej.annotation.Nullable;
import ej.microui.display.Font;

/**
 * A font profile represents a font with a family, a size and a style.
 */
public class FontProfile {

	/**
	 * Font sizes.
	 */
	public enum FontSize {
		/**
		 * Extra extra small.
		 */
		XX_SMALL,
		/**
		 * Extra small.
		 */
		X_SMALL,
		/**
		 * Small.
		 */
		SMALL,
		/**
		 * Medium.
		 */
		MEDIUM,
		/**
		 * Large.
		 */
		LARGE,
		/**
		 * Extra large.
		 */
		X_LARGE,
		/**
		 * Extra extra large.
		 */
		XX_LARGE,
		/**
		 * Manual fixed size length.
		 */
		LENGTH/* custom */;
	}

	private String family;
	private FontSize size;
	private int sizeValue; // used only for LENGTH & PERCENT
	private int style; // MicroUI Font styles

	/**
	 * Creates a font profile.
	 */
	public FontProfile() {
		this.family = ""; //$NON-NLS-1$
		this.size = FontSize.MEDIUM;
		this.style = Font.STYLE_PLAIN;
	}

	/**
	 * Creates a font profile specifying its attributes.
	 *
	 * @param family
	 *            the font family to set.
	 * @param size
	 *            the size to set.
	 * @param style
	 *            the style to set.
	 * @throws IllegalArgumentException
	 *             if the given size is {@link FontSize#LENGTH}.
	 */
	public FontProfile(String family, FontSize size, int style) {
		if (size == FontSize.LENGTH) {
			throw new IllegalArgumentException();
		}
		this.family = family;
		this.size = size;
		this.style = style;
	}

	/**
	 * Creates a font profile specifying its attributes.
	 * <p>
	 * The size {@link FontSize#LENGTH} is used associated with the given size in pixels.
	 *
	 * @param family
	 *            the font family to set.
	 * @param size
	 *            the size in pixels.
	 * @param style
	 *            the style to set.
	 */
	public FontProfile(String family, int size, int style) {
		this.family = family;
		this.size = FontSize.LENGTH;
		this.sizeValue = size;
		this.style = style;
	}

	/**
	 * Gets the font family.
	 *
	 * @return the family.
	 */
	public String getFamily() {
		return this.family;
	}

	/**
	 * Sets the family. We assume a font belong to a font family if its name starts with this font family. The name of a
	 * font is accessible with {@link Font#getDescriptor()}.
	 *
	 * @param family
	 *            the family to set.
	 */
	public void setFamily(String family) {
		this.family = family;
	}

	/**
	 * Gets the font size.
	 *
	 * @return the size.
	 */
	public FontSize getSize() {
		return this.size;
	}

	/**
	 * Sets the size.
	 *
	 * @param size
	 *            the size to set.
	 */
	public void setSize(FontSize size) {
		this.size = size;
	}

	/**
	 * Gets the font size value.
	 * <p>
	 * Used only if size is one of {@link FontSize#LENGTH}.
	 *
	 * @return the size value.
	 */
	public int getSizeValue() {
		return this.sizeValue;
	}

	/**
	 * Sets the size value.
	 *
	 * @param sizeValue
	 *            the size value to set.
	 */
	public void setSizeValue(int sizeValue) {
		this.sizeValue = sizeValue;
	}

	/**
	 * Gets the font style.
	 *
	 * @return the style.
	 */
	public int getStyle() {
		return this.style;
	}

	/**
	 * Sets the style.
	 *
	 * @param style
	 *            the style to set.
	 */
	public void setStyle(int style) {
		this.style = style;
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof FontProfile) {
			FontProfile fontProfile = (FontProfile) obj;
			return this.family.equals(fontProfile.family) && this.size == fontProfile.size
					&& this.sizeValue == fontProfile.sizeValue && this.style == fontProfile.style;
		}
		return false;
	}

	@Override
	public int hashCode() {
		// Since the fields are not final, they can't be used to compute a constant hashcode.
		return 0;
	}

}
