/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 *
 */
package ej.wadapps.storage;

import java.io.IOException;
import java.io.InputStream;

/**
 * Defines the storage mechanism.
 * <p>
 * The storage use an identifier to store, retrieve or remove a data. An identifier is an unlimited-length sequence
 * (greater than 0) of letters, digits and special characters, the first of which must be a letter. The "letters"
 * includes lowercase(a-z), uppercase(A-Z) ASCII latin characters and ASCII underscore(_). The "digits" includes the
 * ASCII digit(0-9). The special characters includes ASCII hyphen(-) and ASCII dot(.).
 *
 */
public interface Storage {
	/**
	 * Stores an input stream that will be referenced with an ID.
	 * <p>
	 * If a data is already stored with this ID, it is overwritten.
	 *
	 * @param id
	 *            the ID of the stored data.
	 * @param stream
	 *            the data to be stored.
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @throws IllegalArgumentException
	 *             if the given ID is not a valid identifier.
	 * @throws NullPointerException
	 *             if one of the given parameters is {@code null}.
	 */
	void store(String id, InputStream stream) throws IOException;

	/**
	 * Returns the data stored with a specific ID or {@code null} if there is none.
	 *
	 * @param id
	 *            the ID of the data to be returned.
	 * @return the data stored with the given ID or {@code null}.
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @throws IllegalArgumentException
	 *             if the given ID is not a valid identifier.
	 * @throws NullPointerException
	 *             if one of the given parameters is {@code null}.
	 */
	InputStream load(String id) throws IOException;

	/**
	 * Changes the ID of a data.
	 * <p>
	 * If a data is already stored with the new ID, it is overwritten.
	 *
	 * @param src
	 *            the old ID of the data.
	 * @param dst
	 *            the new ID of the data.
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @throws IllegalArgumentException
	 *             if one of the given IDs is not a valid identifier.
	 * @throws IllegalArgumentException
	 *             if there is no data stored with the old ID.
	 * @throws NullPointerException
	 *             if one of the given parameters is {@code null}.
	 */
	void move(String src, String dst) throws IOException;

	/**
	 * Removes the data stored with an ID.
	 *
	 * @param id
	 *            the ID of the data to return.
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @throws IllegalArgumentException
	 *             if the given ID is not a valid identifier.
	 * @throws IllegalArgumentException
	 *             if there is no data stored with the given ID.
	 * @throws NullPointerException
	 *             if one of the given parameters is {@code null}.
	 */
	void remove(String id) throws IOException;

	/**
	 * Tests whether a data exists for this ID.
	 *
	 * @param id
	 *            the storage-key to check.
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @return <code>true</code> if the given ID exists, <code>false</code> otherwise
	 */
	boolean exists(String id) throws IOException;

	/**
	 * Returns all IDs of the stored data.
	 *
	 * @return an array of all IDs.
	 * @throws IOException
	 *             if an I/O error occurs.
	 */
	String[] getIds() throws IOException;
}
