/*
 * Java
 *
 * Copyright 2017-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.basic.drawing;

import ej.animation.Animation;
import ej.animation.Animator;
import ej.annotation.Nullable;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.microui.display.GraphicsContext;
import ej.microui.display.shape.AntiAliasedShapes;
import ej.motion.linear.LinearMotion;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.basic.PagingIndicator;

/**
 * The bullet paging indicator is a widget that displays a set of dots following a line. A bigger dot indicates the
 * selected index.
 * <p>
 * The bullet paging indicator is hidden by sliding smoothly the dots outside the drawing area.
 */
public class BulletPagingIndicator extends PagingIndicator implements Animation {

	private static final int HIDE_DURATION = 300;

	private int bulletsPosition;
	@Nullable
	private LinearMotion motion;

	/**
	 * Creates an horizontal bullet paging indicator.
	 */
	public BulletPagingIndicator() {
		super();
	}

	/**
	 * Creates a bullet paging indicator.
	 *
	 * @param horizontal
	 *            <code>true</code> if the paging indicator is horizontal, <code>false</code> otherwise.
	 */
	public BulletPagingIndicator(boolean horizontal) {
		super(horizontal);
	}

	@Override
	public void hideNotify() {
		super.hideNotify();
		cancelHide();
	}

	@Override
	public void show() {
		cancelHide();
		this.bulletsPosition = 0;
	}

	private void cancelHide() {
		Animator animator = ServiceLoaderFactory.getServiceLoader().getService(Animator.class, Animator.class);
		animator.stopAnimation(this);
	}

	@Override
	public void hide() {
		int bulletSize;
		if (isHorizontal()) {
			bulletSize = getHeight();
		} else {
			bulletSize = getWidth();
		}
		this.motion = new LinearMotion(this.bulletsPosition, bulletSize, HIDE_DURATION);
		Animator animator = ServiceLoaderFactory.getServiceLoader().getService(Animator.class, Animator.class);
		animator.startAnimation(this);
	}

	@Override
	public boolean tick(long currentTimeMillis) {
		// Cursors showing/hiding tick.
		LinearMotion motion = this.motion;
		if (motion == null) {
			throw new NullPointerException();
		}
		this.bulletsPosition = motion.getCurrentValue();
		repaint();
		return !motion.isFinished();
	}

	@Override
	public void renderContent(GraphicsContext g, Style style, Rectangle bounds) {
		int boundsWidth = bounds.getWidth();
		int boundsHeight = bounds.getHeight();

		g.setColor(style.getForegroundColor());

		boolean horizontal = isHorizontal();
		int itemsCount = getItemsCount();
		if (itemsCount == 0) {
			// Nothing to draw.
			return;
		}

		int currentItemIndex = getSelectedItem();
		int nextItemIndex = (currentItemIndex + 1) % itemsCount;

		// Compute sizes and initial position.
		int bigSize;
		int x;
		int y;
		if (horizontal) {
			bigSize = boundsHeight - 1; // Minus 1 for antialiasing.
			x = (boundsWidth - bigSize * itemsCount) / 2;
			y = boundsHeight / 2 + this.bulletsPosition;
		} else {
			bigSize = boundsWidth - 1; // Minus 1 for antialiasing.
			x = boundsWidth / 2 + this.bulletsPosition;
			y = (boundsHeight - bigSize * itemsCount) / 2;
		}
		int smallSize = bigSize / 2;
		int percentSize = (int) (smallSize * getPercent());
		int nextSize = bigSize - percentSize;
		int currentSize = smallSize + percentSize;
		// Draw dots.
		for (int i = -1; ++i < itemsCount;) {
			int size;
			if (i == currentItemIndex) {
				size = currentSize;
			} else if (i == nextItemIndex) {
				size = nextSize;
			} else {
				size = smallSize;
			}
			AntiAliasedShapes.Singleton.setThickness(size);

			if (horizontal) {
				AntiAliasedShapes.Singleton.drawPoint(g, x + bigSize / 2, y);
				x += bigSize;
			} else {
				AntiAliasedShapes.Singleton.drawPoint(g, x, y + bigSize / 2);
				y += bigSize;
			}
		}
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int baseSize = StyleHelper.getFont(style).getHeight();
		if ((baseSize & 0x1) == 1) {
			// Make this size event to make sure the half is exactly the half!
			baseSize++;
		}
		// Add 1 for antialiasing for every dimension.
		int size = baseSize * getItemsCount() + 1;
		if (isHorizontal()) {
			return new Rectangle(0, 0, size, baseSize + 1);
		} else {
			return new Rectangle(0, 0, baseSize + 1, size);
		}
	}

}
