/*
 * Java
 *
 * Copyright 2009-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.basic.drawing;

import ej.annotation.Nullable;
import ej.bon.XMath;
import ej.microui.display.GraphicsContext;
import ej.microui.display.shape.AntiAliasedShapes;
import ej.motion.util.MotionAnimator;
import ej.style.Style;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.StyledWidget;

/**
 * A scrollbar holds the current step of a scrolling element.
 * <p>
 * The size of a scrollbar is dependent on the font size.
 * <p>
 * This example shows a simple scrollbar:
 *
 * <pre>
 * Scrollbar scrollbar = new Scrollbar(100);
 * scrollbar.setHorizontal(false);
 * scrollbar.show();
 * scrollbar.setValue(50);
 * </pre>
 *
 * <img src="../doc-files/scrollbar-simple.png" alt="Simple scrollbar.">
 * <p>
 * This example shows a styled scrollbar:
 *
 * <pre>
 * Scrollbar scrollbar = new Scrollbar(100);
 * scrollbar.show();
 * scrollbar.setHorizontal(false);
 * scrollbar.setValue(50);
 *
 * EditableStyle scrollbarStyle = new EditableStyle();
 * scrollbarStyle.setForegroundColor(Colors.BLUE);
 * scrollbarStyle.setBackgroundColor(Colors.SILVER);
 * scrollbarStyle.setPadding(new SimpleOutline(1));
 * scrollbar.mergeStyle(scrollbarStyle);
 * </pre>
 *
 * <img src="../doc-files/scrollbar-styled.png" alt="Styled scrollbar.">
 */
public class Scrollbar extends StyledWidget {

	private static final int MINIMUM_RATIO = 4;
	private static final int MINIMUM_SIZE = 4;

	// private static final int ANIMATION_DELAY = 300;

	private boolean horizontal;
	private int maximum;
	private int value;
	private int requiredValue;
	private int visibilityLevel;
	@Nullable
	private MotionAnimator visiblityAnimator;
	// private boolean hidden;

	/**
	 * Creates a new scrollbar with the specified size.
	 *
	 * @param maximum
	 *            the maximum bound.
	 */
	public Scrollbar(int maximum) {
		this.maximum = Math.max(maximum, 0);
		// this.hidden = true;
		internalSetVisibilityLevel(GraphicsContext.TRANSPARENT);
	}

	/**
	 * Sets the scrollbar orientation: horizontal or vertical.
	 *
	 * @param horizontal
	 *            <code>true</code> to set the slider horizontal, <code>false</code> to set the slider vertical.
	 */
	public void setHorizontal(boolean horizontal) {
		this.horizontal = horizontal;
		invalidate();
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int size = getSize(style);
		int preferredWidth;
		int preferredHeight;
		// The length of the scrollbar is minimum MINIMUM_RATIO times bigger than the thickness.
		// The length of the scrollbar area is twice the scrollbar length.
		// Add 1 for the fade.
		int length = (size * MINIMUM_RATIO) << 1;
		if (this.horizontal) {
			preferredWidth = length;
			preferredHeight = size;
		} else {
			preferredWidth = size;
			preferredHeight = length;
		}
		return new Rectangle(0, 0, preferredWidth + 1, preferredHeight + 1);
	}

	@Override
	public void renderContent(GraphicsContext g, Style style, Rectangle bounds) {
		if (this.visibilityLevel == GraphicsContext.TRANSPARENT) {
			// nothing to paint
			return;
		}

		// Remove 1 for the fade.
		int width = bounds.getWidth() - 1;
		int height = bounds.getHeight() - 1;

		int alignment = style.getAlignment();

		int size = getSize(style);

		g.setColor(style.getForegroundColor());
		AntiAliasedShapes.Singleton.setThickness(size - 1);
		AntiAliasedShapes.Singleton.setFade(1);
		int barX;
		int barY;
		int barX2;
		int barY2;
		if (this.horizontal) {
			barY = AlignmentHelper.computeYTopCorner(size, 0, height, alignment) + (size >> 1);
			int barWidth = getBarLength(width, size);
			barX = getBarPosition(width, barWidth, size);
			barX2 = barX + barWidth;
			barY2 = barY;
		} else {
			barX = AlignmentHelper.computeXLeftCorner(size, 0, width, alignment) + (size >> 1);
			int barHeight = getBarLength(height, size);
			barY = getBarPosition(height, barHeight, size);
			barX2 = barX;
			barY2 = barY + barHeight;
		}
		AntiAliasedShapes.Singleton.drawLine(g, barX, barY, barX2, barY2);
	}

	private int getBarLength(int availableLength, int size) {
		availableLength = availableLength - (size << 1);
		int excess = Math.abs(this.requiredValue - this.value) >> 1;
		int expectedLength = Math.max(availableLength * availableLength / this.maximum, size * MINIMUM_RATIO);
		// Don't know why the minimum size is '2', but with lower values the result is not correct.
		int finalLength = XMath.limit(expectedLength - excess, 2, availableLength - excess);
		return finalLength - 1;
	}

	private int getBarPosition(int totalSize, int barSize, int size) {
		return this.value * (totalSize - barSize - size) / this.maximum + (size >> 1);
	}

	private int getSize(Style style) {
		return Math.max(StyleHelper.getFont(style).getHeight() >> 2, MINIMUM_SIZE);
	}

	/**
	 * Hides the scrollbar.
	 */
	public void hide() {
		// if (this.hidden) {
		internalSetVisibilityLevel(GraphicsContext.TRANSPARENT);
		// } else {
		// Motion visibilityMotion = new LinearMotion(GraphicsContext.OPAQUE, GraphicsContext.TRANSPARENT,
		// ANIMATION_DELAY);
		// this.visiblityAnimator = new MotionAnimator(visibilityMotion, new MotionListenerAdapter() {
		// @Override
		// public void step(int value) {
		// internalSetVisibilityLevel(value);
		// }
		// });
		// Timer timer = ServiceLoaderFactory.getServiceLoader().getService(Timer.class, Timer.class);
		// this.visiblityAnimator.start(timer, ANIMATION_PERIOD);
		// }
	}

	@Override
	public void showNotify() {
		// this.hidden = false;
	}

	@Override
	public void hideNotify() {
		// this.hidden = true;
		hide();
		stopAnimator();
	}

	private void stopAnimator() {
		MotionAnimator visiblityAnimator = this.visiblityAnimator;
		if (visiblityAnimator != null) {
			visiblityAnimator.stop();
		}
	}

	/**
	 * Shows the scrollbar.
	 */
	public void show() {
		if (this.visibilityLevel != GraphicsContext.OPAQUE) {
			stopAnimator();
			internalSetVisibilityLevel(GraphicsContext.OPAQUE);
		}
	}

	private void internalSetVisibilityLevel(int visibilityLevel) {
		visibilityLevel = XMath.limit(visibilityLevel, GraphicsContext.TRANSPARENT, GraphicsContext.OPAQUE);
		if (visibilityLevel != this.visibilityLevel) {
			this.visibilityLevel = visibilityLevel;
			repaint();
		}
	}

	/**
	 * From {@link GraphicsContext#TRANSPARENT} (hidden) to {@link GraphicsContext#OPAQUE} (fully visible).
	 *
	 * @return the visibilityLevel
	 */
	public int getVisibilityLevel() {
		return this.visibilityLevel;
	}

	/**
	 * @return <code>true</code> if the scrollbar is horizontal, <code>false</code> otherwise.
	 */
	public boolean isHorizontal() {
		return this.horizontal;
	}

	/**
	 * Gets the max step.
	 *
	 * @return the max step.
	 */
	public int getMaximum() {
		return this.maximum;
	}

	/**
	 * Gets the current value.
	 *
	 * @return the current value.
	 */
	public int getValue() {
		return this.value;
	}

	/**
	 * Updates the current value with an increment.
	 *
	 * @param increment
	 *            the increment to apply to the current value.
	 */
	public void updateValue(int increment) {
		setValue(this.requiredValue + increment);
	}

	/**
	 * Sets the current value.
	 * <p>
	 * The scrollbar is not repainted. The caller is responsible for the repaint.
	 *
	 * @param value
	 *            the value to set.
	 */
	public void setValue(int value) {
		this.requiredValue = value;
		this.value = XMath.limit(value, 0, this.maximum);
		// Do not repaint here. The scroll composite containing the scrollbar is responsible for the repaint. Avoid
		// repainting two times the bar and slowing down the scroll!
	}

	/**
	 * Sets the maximum bound of the scrollbar.
	 *
	 * @param maximum
	 *            the maximum to set.
	 */
	public void setMaximum(int maximum) {
		this.maximum = Math.max(maximum, 0);
	}

}
