/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.basic.drawing;

import ej.microui.display.GraphicsContext;
import ej.microui.display.shape.AntiAliasedShapes;
import ej.style.Style;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.basic.Box;

/**
 * Implementation of a box that represents a switch.
 * <p>
 * The size of a switch box is dependent on the font size.
 * <p>
 * This example shows simple switch boxes (one checked, the other not):
 *
 * <pre>
 * SwitchBox switchBox = new SwitchBox();
 * switchBox.setChecked(true);
 * SwitchBox switchBox2 = new SwitchBox();
 * </pre>
 *
 * <img src="../doc-files/switchbox-simple.png" alt="Switch box.">
 * <p>
 * This example shows a styled switche box:
 *
 * <pre>
 * SwitchBox switchBox = new SwitchBox();
 * switchBox.setChecked(true);
 * SwitchBox switchBox2 = new SwitchBox();
 *
 * Stylesheet stylesheet = StyleHelper.getStylesheet();
 *
 * EditableStyle switchBoxStyle = new EditableStyle();
 * switchBoxStyle.setBorder(new SimpleRoundedBorder(Integer.MAX_VALUE, 2));
 * switchBoxStyle.setForegroundColor(Colors.BLUE);
 * switchBoxStyle.setPadding(new ComplexOutline(4, 20, 4, 4));
 * switchBoxStyle.setAlignment(GraphicsContext.LEFT | GraphicsContext.VCENTER);
 * stylesheet.addRule(new TypeSelector(SwitchBox.class), switchBoxStyle);
 *
 * EditableStyle checkedSwitchboxStyle = new EditableStyle();
 * checkedSwitchboxStyle.setBorderColor(Colors.BLUE);
 * checkedSwitchboxStyle.setPadding(new ComplexOutline(4, 4, 4, 20));
 * checkedSwitchboxStyle.setAlignment(GraphicsContext.RIGHT | GraphicsContext.VCENTER);
 * stylesheet.addRule(new AndCombinator(new TypeSelector(SwitchBox.class), new StateSelector(State.Checked)),
 * 		checkedSwitchboxStyle);
 * </pre>
 *
 * <img src="../doc-files/switchbox-styled-checked.png" alt="Styled switch box.">
 */
public class SwitchBox extends Box {

	private static final int THICKNESS = 2;
	private static final int MINIMUM_SIZE = 12;

	@Override
	public void renderContent(GraphicsContext g, Style style, Rectangle bounds) {
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		AntiAliasedShapes antiAliasedShapes = AntiAliasedShapes.Singleton;
		antiAliasedShapes.setThickness(THICKNESS);
		antiAliasedShapes.setFade(1);

		g.setColor(style.getForegroundColor());
		int alignment = style.getAlignment();
		int size = Math.min(width, height) - 2 * THICKNESS;
		int leftX = AlignmentHelper.computeXLeftCorner(size, THICKNESS, width - 2 * THICKNESS, alignment);
		int topY = AlignmentHelper.computeYTopCorner(size, THICKNESS, height - 2 * THICKNESS, alignment);
		if (isChecked()) {
			int centerX = leftX + (size >> 1);
			antiAliasedShapes.drawLine(g, centerX, topY, centerX, topY + size);
		} else {
			antiAliasedShapes.drawCircle(g, leftX, topY, size);
		}
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle availableSize) {
		int size = Math.max(StyleHelper.getFont(style).getHeight() >> 1, MINIMUM_SIZE);
		if ((size & 0x1) == 0) {
			// Forces it to be odd to center the content.
			size++;
		}
		return new Rectangle(0, 0, size, size);
	}

}
