/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.basic.image;

import ej.microui.display.GraphicsContext;
import ej.microui.display.Image;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.basic.drawing.Slider;
import ej.widget.model.BoundedRangeModel;
import ej.widget.model.DefaultBoundedRangeModel;

/**
 * A slider using some images to render.
 */
public class ImageSlider extends Slider {

	/**
	 * Creates a horizontal slider with the given model.
	 *
	 * @param model
	 *            the model to use.
	 */
	public ImageSlider(BoundedRangeModel model) {
		super(model);
	}

	/**
	 * Creates a horizontal slider with a default bounded range model.
	 *
	 * @param min
	 *            the minimum value of the slider.
	 * @param max
	 *            the maximum value of the slider.
	 * @param initialValue
	 *            the initial value of the slider.
	 * @see DefaultBoundedRangeModel
	 */
	public ImageSlider(int min, int max, int initialValue) {
		super(min, max, initialValue);
	}

	@Override
	public void renderContent(GraphicsContext g, Style style, Rectangle bounds) {
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		int foregroundColor = style.getForegroundColor();
		g.setColor(foregroundColor);

		float percentComplete = getPercentComplete();
		Image bar;
		int xCursor;
		int yCursor;

		if (this.horizontal) {
			bar = getBarImage(true);
			int barWidth = bar.getWidth();
			int xBarStart = (width - barWidth) >> 1;
			xCursor = xBarStart + (int) (barWidth * percentComplete);
			yCursor = height >> 1;
		} else {
			bar = getBarImage(false);
			int barHeight = bar.getHeight();
			int yBarStart = (height - barHeight) >> 1;
			xCursor = width >> 1;
			yCursor = yBarStart + (int) (barHeight * percentComplete);
		}

		// Draws the bar.
		g.drawImage(bar, width >> 1, height >> 1, GraphicsContext.HCENTER | GraphicsContext.VCENTER);

		// Draws the cursor.
		g.drawImage(getCursorImage(), xCursor, yCursor, GraphicsContext.HCENTER | GraphicsContext.VCENTER);
	}

	@Override
	protected float computePercentComplete(int pointerX, int pointerY) {
		Rectangle contentBounds = getContentBounds();
		// The size of the bar consider the size of the cursor (room left on both sides).
		if (this.horizontal) {
			int barWidth = getBarImage(true).getWidth();
			return (float) (getRelativeX(pointerX) - contentBounds.getX()
					- ((contentBounds.getWidth() - barWidth) >> 1)) / barWidth;
		} else {
			int barHeight = getBarImage(false).getHeight();
			return (float) (getRelativeY(pointerY) - contentBounds.getY()
					- ((contentBounds.getHeight() - barHeight) >> 1)) / barHeight;
		}
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle availableSize) {
		int width;
		int height;
		Image cursor = getCursorImage();

		if (this.horizontal) {
			Image bar = getBarImage(true);
			width = bar.getWidth() + cursor.getWidth();
			height = Math.max(bar.getHeight(), cursor.getHeight());
		} else {
			Image bar = getBarImage(false);
			width = Math.max(bar.getWidth(), cursor.getWidth());
			height = bar.getHeight() + cursor.getHeight();
		}

		return new Rectangle(0, 0, width, height);
	}

	private Image getBarImage(boolean horizontal) {
		String imagePath = horizontal ? ImageHelper.getTheme().getSliderHorizontalBarPath()
				: ImageHelper.getTheme().getSliderVerticalBarPath();
		Image image = StyleHelper.getImage(imagePath);
		if (image == null) {
			throw new NullPointerException();
		}
		return image;
	}

	private Image getCursorImage() {
		String path = ImageHelper.getTheme().getSliderCursorPath();
		Image image = StyleHelper.getImage(path);
		if (image == null) {
			throw new NullPointerException();
		}
		return image;
	}

}
