/*
 * Java
 *
 * Copyright 2017-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.container;

import java.util.Iterator;

import ej.annotation.Nullable;
import ej.bon.Timer;
import ej.bon.TimerTask;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.mwt.MWT;
import ej.mwt.Widget;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.widget.StyledComposite;
import ej.widget.animation.AnimationListener;
import ej.widget.basic.PagingIndicator;
import ej.widget.util.SwipeEventHandler;
import ej.widget.util.Swipeable;

/**
 * The abstract carousel is a composite that contains a list of widget. It defines the framework to create any kind of
 * carousel.
 * <p>
 * If the carousel is cyclic, when the last element is reached, the next one is the first one, and the other way round.
 * <p>
 * The carousel also displays a cursor showing the selected item. This cursor can be hidden automatically.
 *
 * @since 2.3.0
 */
public abstract class AbstractCarousel extends StyledComposite implements Swipeable, AnimationListener {

	private static final int HIDE_DELAY = 500;

	/**
	 * Whether the carousel is horizontal or not.
	 */
	protected final boolean horizontal;
	/**
	 * Whether the carousel is cyclic or not.
	 */
	protected final boolean cyclic;
	/**
	 * Whether the cursor is hidden automatically or not.
	 */
	protected final boolean autoHide;
	/**
	 * The cursor.
	 */
	@Nullable
	protected final PagingIndicator cursor;

	@Nullable
	private SwipeEventHandler swipeEventHandler;

	@Nullable
	private TimerTask hideTask;
	private int selectedIndex;

	/**
	 * Creates a carousel without cursor.
	 *
	 * @param horizontal
	 *            <code>true</code> if the carousel is horizontal, <code>false</code> otherwise.
	 * @param cyclic
	 *            <code>true</code> if the carousel is cyclic, <code>false</code> otherwise.
	 */
	public AbstractCarousel(boolean horizontal, boolean cyclic) {
		this(horizontal, cyclic, null, false);
	}

	/**
	 * Creates a carousel with a cursor that can be hidden automatically or not.
	 *
	 * @param horizontal
	 *            <code>true</code> if the carousel is horizontal, <code>false</code> otherwise.
	 * @param cyclic
	 *            <code>true</code> if the carousel is cyclic, <code>false</code> otherwise.
	 * @param cursor
	 *            the cursor that indicate the progress in the carousel.
	 * @param autoHideCursor
	 *            <code>true</code> if the cursor hides automatically, <code>false</code> otherwise.
	 */
	public AbstractCarousel(boolean horizontal, boolean cyclic, @Nullable PagingIndicator cursor,
			boolean autoHideCursor) {
		this.horizontal = horizontal;
		this.cyclic = cyclic;
		this.cursor = cursor;
		if (cursor != null) {
			cursor.setHorizontal(horizontal);
			// Adds the cursor here to make sure that a method that needs to count the widgets (getWidgetsCount() - 1)
			// before validation get the right value!
			add(cursor);
		} else {
			autoHideCursor = false;
		}
		this.autoHide = autoHideCursor;
	}

	@Override
	public void showNotify() {
		super.showNotify();

		PagingIndicator cursor = this.cursor;
		// Put focus on the first widget.
		Iterator<Widget> iterator = iterator();
		// Each child takes the full width and the height it needs.
		while (iterator.hasNext()) {
			Widget widget = iterator.next();
			if (widget != cursor) {
				widget.requestFocus();
				break;
			}
		}

		if (this.autoHide) {
			hideCursorsAsynchronous();
		}
	}

	@Override
	public void hideNotify() {
		super.hideNotify();

		SwipeEventHandler swipeEventHandler = this.swipeEventHandler;
		if (swipeEventHandler != null) {
			swipeEventHandler.stop();
		}

		cancelHideCursor();
	}

	@Override
	public void add(Widget widget) throws NullPointerException, IllegalArgumentException {
		super.add(widget);
	}

	@Override
	public void remove(Widget widget) throws NullPointerException {
		super.remove(widget);
	}

	@Override
	public void removeAllWidgets() {
		super.removeAllWidgets();
	}

	/**
	 * Sets the currently selected widget index.
	 *
	 * @param selectedIndex
	 *            the index to set.
	 */
	protected void setSelectedIndex(int selectedIndex) {
		this.selectedIndex = selectedIndex;
	}

	/**
	 * Gets the currently selected widget index.
	 *
	 * @return the selected widget index.
	 */
	public int getSelectedIndex() {
		return this.selectedIndex;
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		boolean horizontal = this.horizontal;

		int boundsWidth = bounds.getWidth();
		int boundsHeight = bounds.getHeight();

		boolean computeWidth = boundsWidth == MWT.NONE;
		boolean computeHeight = boundsHeight == MWT.NONE;

		PagingIndicator cursor = this.cursor;
		boolean autoHide = this.autoHide;
		int cursorWidth;
		int cursorHeight;
		if (cursor != null) {
			// Add cursor here to make sure it's the last one (it is over the other ones when autohide).
			remove(cursor);
			add(cursor);

			// The cursor takes the full width and the height it needs when horizontal, the opposite when vertical.
			cursor.setItemsCount(getWidgetsCount() - 1);
			cursor.validate(MWT.NONE, boundsHeight);
			cursorWidth = cursor.getPreferredWidth();
			cursorHeight = cursor.getPreferredHeight();
			if (!autoHide) {
				// When the cursor is hidden automatically, it is above the list, otherwise it takes some space.
				if (horizontal) {
					if (!computeHeight) {
						boundsHeight -= cursorHeight;
					}
				} else {
					if (!computeWidth) {
						boundsWidth -= cursorWidth;
					}
				}
			}
		} else {
			cursorWidth = 0;
			cursorHeight = 0;
		}

		Rectangle childrenSize = validateChildren(boundsWidth, boundsHeight, horizontal);
		int width = childrenSize.getWidth();
		int height = childrenSize.getHeight();

		if (computeWidth) {
			boundsWidth = width;
		}
		if (computeHeight) {
			boundsHeight = height;
		}
		// If cursor is null, its width and height are set to 0, do not need to test it.
		if (!autoHide) {
			if (horizontal) {
				boundsHeight += cursorHeight;
			} else {
				boundsWidth += cursorWidth;
			}
		}

		return new Rectangle(0, 0, boundsWidth, boundsHeight);
	}

	/**
	 * Lays out the children.
	 *
	 * @param boundsWidth
	 *            the width hint.
	 * @param boundsHeight
	 *            the height hint.
	 * @param horizontal
	 *            <code>true</code> if the carousel is horizontal, <code>false</code> otherwise.
	 * @return the preferred size of the children.
	 */
	protected abstract Rectangle validateChildren(int boundsWidth, int boundsHeight, boolean horizontal);

	@Override
	protected void setBoundsContent(Rectangle bounds) {
		int boundsX = bounds.getX();
		int boundsY = bounds.getY();
		int boundsWidth = bounds.getWidth();
		int boundsHeight = bounds.getHeight();

		boolean horizontal = this.horizontal;

		// The cursor takes the full width and the height it needs when horizontal, the opposite when vertical.
		PagingIndicator cursor = this.cursor;
		boolean autoHide = this.autoHide;
		if (cursor != null) {
			int cursorWidth = cursor.getPreferredWidth();
			int cursorHeight = cursor.getPreferredHeight();
			if (horizontal) {
				cursor.setBounds(boundsX, boundsY + boundsHeight - cursorHeight, boundsWidth, cursorHeight);
			} else {
				cursor.setBounds(boundsX + boundsWidth - cursorWidth, boundsY, cursorWidth, boundsHeight);
			}

			if (!autoHide) {
				if (horizontal) {
					boundsHeight -= cursorHeight;
				} else {
					boundsWidth -= cursorWidth;
				}
			}
		}

		int widgetsCount = getWidgetsCount();
		if (cursor != null) {
			widgetsCount--;
		}
		this.selectedIndex = Math.min(this.selectedIndex, widgetsCount - 1);

		if (widgetsCount > 0) {
			this.swipeEventHandler = setBoundsChildren(bounds, widgetsCount);
			if (autoHide) {
				this.swipeEventHandler.setAnimationListener(this);
			}

			if (this.selectedIndex != -1) {
				goTo(this.selectedIndex);
			}
		} else {
			this.swipeEventHandler = null;
		}
	}

	/**
	 * Sets the bounds of the children.
	 *
	 * @param bounds
	 *            the bounds of this carousel.
	 * @param widgetsCount
	 *            the number of widgets in the carousel.
	 * @return the swipe event handler to scroll the widgets.
	 */
	protected abstract SwipeEventHandler setBoundsChildren(Rectangle bounds, int widgetsCount);

	/**
	 * Moves a widget when scrolling.
	 *
	 * @param widget
	 *            the widget to move.
	 * @param horizontal
	 *            <code>true</code> if the carousel is horizontal, <code>false</code> otherwise.
	 * @param totalSize
	 *            the cumulated size of the children (width if horizontal, height if vertical).
	 * @param cyclic
	 *            <code>true</code> if the carousel is cyclic, <code>false</code> otherwise.
	 * @param shift
	 *            the move shift.
	 * @return the size of the widget (width if horizontal, height if vertical).
	 */
	protected int moveWidget(Widget widget, boolean horizontal, int totalSize, boolean cyclic, int shift) {
		int widgetX = widget.getX();
		int widgetY = widget.getY();
		int widgetSize;
		if (horizontal) {
			widgetX -= shift;
			widgetSize = widget.getWidth();
		} else {
			widgetY -= shift;
			widgetSize = widget.getHeight();
		}
		if (cyclic) {
			// If cyclic, items need to be moved at the end or at the beginning.
			if (shift > 0) {
				if (horizontal) {
					if (widgetX + widgetSize < 0) {
						// Push at the end.
						widgetX += totalSize;
					}
				} else {
					if (widgetY + widgetSize < 0) {
						// Push at the end.
						widgetY += totalSize;
					}
				}
			} else {
				if (horizontal) {
					if (widgetX > getWidth()) {
						// Push at the beginning.
						widgetX -= totalSize;
					}
				} else {
					if (widgetY > getHeight()) {
						// Push at the beginning.
						widgetY -= totalSize;
					}
				}
			}
		}
		widget.setLocation(widgetX, widgetY);
		return widgetSize;
	}

	@Override
	public boolean handleEvent(int event) {
		SwipeEventHandler swipeEventHandler = this.swipeEventHandler;
		if (swipeEventHandler != null && swipeEventHandler.handleEvent(event)) {
			return true;
		}
		return super.handleEvent(event);

	}

	/**
	 * Selects the previous widget.
	 */
	public void goToPrevious() {
		goToIncrement(false, 0);
	}

	/**
	 * Selects the next widget.
	 */
	public void goToNext() {
		goToIncrement(true, 0);
	}

	/**
	 * Animates the selection of the previous widget.
	 *
	 * @param duration
	 *            the duration of the animation.
	 * @throws IllegalArgumentException
	 *             if the given duration is less or equal than <code>0</code>.
	 */
	public void goToPrevious(int duration) {
		if (duration <= 0) {
			throw new IllegalArgumentException();
		}
		goToIncrement(false, duration);
	}

	/**
	 * Animates the selection of the next widget.
	 *
	 * @param duration
	 *            the duration of the animation.
	 * @throws IllegalArgumentException
	 *             if the given duration is less or equal than <code>0</code>.
	 */
	public void goToNext(int duration) {
		if (duration <= 0) {
			throw new IllegalArgumentException();
		}
		goToIncrement(true, duration);
	}

	private void goToIncrement(boolean next, int duration) {
		int increment = next ? 1 : -1;
		int nextItem = this.selectedIndex + increment;
		int widgetsCount = getWidgetsCount();
		if (this.cursor != null) {
			widgetsCount--;
		}
		if (next) {
			if (nextItem == widgetsCount) {
				if (this.cyclic) {
					nextItem = 0;
				} else {
					// Cannot go downward.
					return;
				}
			}
		} else {
			if (nextItem == -1) {
				if (this.cyclic) {
					nextItem = getWidgetsCount() - 1;
					if (this.cursor != null) {
						nextItem--;
					}
				} else {
					// Cannot go upward.
					return;
				}
			}
		}
		goToInternal(nextItem, duration);
	}

	/**
	 * Selects a widget from its index.
	 *
	 * @param index
	 *            the widget index.
	 * @throws IllegalArgumentException
	 *             if the given index is not valid (between <code>0</code> and the number of items in the carousel).
	 */
	public void goTo(int index) {
		goToInternal(index, 0);
	}

	/**
	 * Animates the selection of a widget from its index.
	 *
	 * @param index
	 *            the widget index.
	 * @param duration
	 *            the duration of the animation.
	 * @throws IllegalArgumentException
	 *             if the given duration is less or equal than <code>0</code>.
	 * @throws IllegalArgumentException
	 *             if the given index is not valid (between <code>0</code> and the number of items in the carousel).
	 */
	public void goTo(int index, long duration) {
		if (duration <= 0) {
			throw new IllegalArgumentException();
		}
		goToInternal(index, duration);
	}

	private void goToInternal(int index, long duration) {
		int widgetsCount = getWidgetsCount();
		if (this.cursor != null) {
			widgetsCount--;
		}
		if (index < 0 || index >= widgetsCount) {
			throw new IllegalArgumentException();
		}

		SwipeEventHandler swipeEventHandler = this.swipeEventHandler;
		if (swipeEventHandler == null) {
			this.selectedIndex = index;
			return;
		}

		if (this.autoHide) {
			showCursor();
			hideCursorsAsynchronous();
		}
		boolean horizontal = this.horizontal;
		PagingIndicator cursor = this.cursor;
		int size = 0;
		int currentIndex = 0;
		Iterator<Widget> iterator = iterator();
		while (iterator.hasNext()) {
			Widget widget = iterator.next();
			if (widget != cursor) {
				if (currentIndex++ == index) {
					if (duration == 0) {
						swipeEventHandler.moveTo(size);
					} else {
						swipeEventHandler.moveTo(size, duration);
					}
					break;
				}
				if (horizontal) {
					size += widget.getWidth();
				} else {
					size += widget.getHeight();
				}
			}
		}
	}

	@Override
	public void onStartAnimation() {
		showCursor();
	}

	@Override
	public void onStopAnimation() {
		hideCursorsAsynchronous();
	}

	private void showCursor() {
		cancelHideCursor();
		final PagingIndicator cursor = this.cursor;
		if (cursor != null) {
			cursor.show();
		}
	}

	/**
	 * Schedules a task to hide the cursor.
	 */
	private void hideCursorsAsynchronous() {
		Timer timer = ServiceLoaderFactory.getServiceLoader().getService(Timer.class);
		this.hideTask = new TimerTask() {
			@Override
			public void run() {
				hideCursor();
			}
		};
		timer.schedule(this.hideTask, HIDE_DELAY);
	}

	/**
	 * Cancels the task that hides the cursor.
	 */
	private void cancelHideCursor() {
		TimerTask hideTask = this.hideTask;
		if (hideTask != null) {
			hideTask.cancel();
		}
	}

	private void hideCursor() {
		final PagingIndicator cursor = this.cursor;
		if (cursor != null) {
			cursor.hide();
		}
	}

}
