/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.container;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import ej.giml.annotation.Element;
import ej.giml.annotation.ElementChild;
import ej.giml.annotation.ElementConstructor;
import ej.mwt.MWT;
import ej.mwt.Widget;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.widget.StyledComposite;

/**
 * A canvas contains widgets that can be freely arranged.
 */
@Element
public class Canvas extends StyledComposite {

	private final Map<Widget, Rectangle> widgetBounds;

	/**
	 * Creates a canvas.
	 */
	@ElementConstructor
	public Canvas() {
		this.widgetBounds = new HashMap<Widget, Rectangle>(1);
	}

	/**
	 * Adds a widget in this composite. The widget is laid out with the given bounds.
	 * <p>
	 * If the given width or height is {@link MWT#NONE}, the widget size (width or/and height) is fitted to its
	 * preferred size.
	 *
	 * @param widget
	 *            the widget to add.
	 * @param x
	 *            the x coordinate of the widget.
	 * @param y
	 *            the y coordinate of the widget.
	 * @param width
	 *            the width of the widget.
	 * @param height
	 *            the height of the widget.
	 * @throws NullPointerException
	 *             if the specified widget is <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the specified widget or one of its children is already connected to a panel.
	 * @see #add(Widget)
	 */
	@ElementChild(cardinality = Integer.MAX_VALUE, ordered = true)
	public void add(Widget widget, int x, int y, int width, int height)
			throws NullPointerException, IllegalArgumentException {
		super.add(widget);
		this.widgetBounds.put(widget, new Rectangle(x, y, width, height));
	}

	@Override
	public void remove(Widget widget) throws NullPointerException {
		super.remove(widget);
		this.widgetBounds.remove(widget);
	}

	@Override
	public void removeAllWidgets() {
		super.removeAllWidgets();
		this.widgetBounds.clear();
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int widthHint = bounds.getWidth();
		int heightHint = bounds.getHeight();

		Map<Widget, Rectangle> widgetBounds = this.widgetBounds;

		int maxWidth = 0;
		int maxHeight = 0;
		Iterator<Widget> iterator = iterator();
		while (iterator.hasNext()) {
			Widget widget = iterator.next();
			Rectangle rectangle = widgetBounds.get(widget);
			int widgetWidthHint = rectangle.getWidth();
			int widgetHeightHint = rectangle.getHeight();
			widget.validate(widgetWidthHint, widgetHeightHint);
			int widgetX = rectangle.getX();
			int widgetY = rectangle.getY();
			int widgetWidth = widgetWidthHint == MWT.NONE ? widget.getPreferredWidth() : widgetWidthHint;
			int widgetHeight = widgetHeightHint == MWT.NONE ? widget.getPreferredHeight() : widgetHeightHint;
			widget.setBounds(widgetX, widgetY, widgetWidth, widgetHeight);

			// Compute biggest size.
			maxWidth = Math.max(maxWidth, widgetX + widgetWidth);
			maxHeight = Math.max(maxHeight, widgetY + widgetHeight);
		}

		if (widthHint == MWT.NONE) {
			widthHint = maxWidth;
		}
		if (heightHint == MWT.NONE) {
			heightHint = maxHeight;
		}

		// Set composite preferred size.
		return new Rectangle(0, 0, widthHint, heightHint);
	}

	@Override
	protected void setBoundsContent(Rectangle bounds) {
		// Nothing to do.
	}

}
