/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.container;

import java.util.ArrayList;
import java.util.Iterator;

import ej.giml.annotation.Element;
import ej.giml.annotation.ElementAttribute;
import ej.giml.annotation.ElementChild;
import ej.giml.annotation.ElementConstructor;
import ej.mwt.MWT;
import ej.mwt.Widget;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.widget.StyledComposite;

/**
 * Lays out its children in a list.
 * <p>
 * All lines (or columns if vertical) can have a different height (or width).
 * <p>
 * Horizontal:
 * <p>
 * <img src="doc-files/list_horizontal.png" alt="Horizontal list.">
 * <p>
 * Vertical:
 * <p>
 * <img src="doc-files/list_vertical.png" alt="Vertical list.">
 */
@Element
public class List extends StyledComposite implements Scrollable {

	private boolean horizontal;
	private final java.util.List<Widget> childrenList;
	private int lastFirstVisible;
	private int totalWidth;
	private int totalHeight;

	/**
	 * Creates a horizontal list composite.
	 */
	public List() {
		this(true);
	}

	/**
	 * Creates a list composite specifying its orientation.
	 *
	 * @param horizontal
	 *            <code>true</code> to set the composite horizontal, <code>false</code> to set the composite vertical.
	 */
	@ElementConstructor
	public List(@ElementAttribute(defaultValue = "true") boolean horizontal) {
		this.horizontal = horizontal;
		this.childrenList = new ArrayList<>(2);
	}

	/**
	 * Sets the composite orientation: horizontal or vertical.
	 *
	 * @param horizontal
	 *            <code>true</code> to set the composite horizontal, <code>false</code> to set the composite vertical.
	 */
	public void setHorizontal(boolean horizontal) {
		this.horizontal = horizontal;
		invalidate();
	}

	@Override
	@ElementChild(cardinality = Integer.MAX_VALUE, ordered = true)
	public void add(Widget widget) throws NullPointerException, IllegalArgumentException {
		this.childrenList.add(widget);
		super.add(widget);
	}

	@Override
	public void remove(Widget widget) {
		this.childrenList.remove(widget);
		super.remove(widget);
	}

	@Override
	public void removeAllWidgets() {
		this.childrenList.clear();
		super.removeAllWidgets();
	}

	@Override
	public Widget getWidget(int index) throws IndexOutOfBoundsException {
		return this.childrenList.get(index);
	}

	@Override
	public Widget[] getWidgets() {
		return this.childrenList.toArray(new Widget[this.childrenList.size()]);
	}

	@Override
	public int getWidgetsCount() {
		return this.childrenList.size();
	}

	@Override
	public Iterator<Widget> iterator() {
		return this.childrenList.iterator();
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int widthHint = bounds.getWidth();
		int heightHint = bounds.getHeight();

		int length = getWidgetsCount();
		if (length != 0) {
			// Compute widgets expected size.
			boolean horizontalLocal = this.horizontal;

			int widgetWidthHint;
			int widgetHeightHint;
			if (horizontalLocal) {
				widgetWidthHint = MWT.NONE;
				widgetHeightHint = heightHint;
			} else {
				widgetWidthHint = widthHint;
				widgetHeightHint = MWT.NONE;
			}
			int totalCellWidth = 0;
			int totalCellHeight = 0;

			Iterator<Widget> iterator = iterator();
			while (iterator.hasNext()) {
				Widget widget = iterator.next();
				widget.validate(widgetWidthHint, widgetHeightHint);

				// Compute total size.
				if (horizontalLocal) {
					totalCellWidth += widget.getPreferredWidth();
					totalCellHeight = Math.max(totalCellHeight, widget.getPreferredHeight());
				} else {
					totalCellWidth = Math.max(totalCellWidth, widget.getPreferredWidth());
					totalCellHeight += widget.getPreferredHeight();
				}
			}

			this.totalWidth = totalCellWidth;
			this.totalHeight = totalCellHeight;

			if (widthHint == MWT.NONE) {
				widthHint = totalCellWidth;
			}
			if (heightHint == MWT.NONE) {
				heightHint = totalCellHeight;
			}
		}

		// Set composite preferred size.
		return new Rectangle(0, 0, widthHint, heightHint);
	}

	@Override
	protected void setBoundsContent(Rectangle bounds) {
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		int length = getWidgetsCount();
		if (length == 0) {
			return;
		}

		float xRatio = (float) width / this.totalWidth;
		float yRatio = (float) height / this.totalHeight;

		// Compute widgets bounds.
		boolean horizontalLocal = this.horizontal;

		float currentX = bounds.getX();
		float currentY = bounds.getY();
		Iterator<Widget> iterator = iterator();
		while (iterator.hasNext()) {
			Widget widget = iterator.next();

			if (horizontalLocal) {
				float widgetWidth = widget.getPreferredWidth() * xRatio;
				int rightX = (int) (currentX + widgetWidth);
				widget.setBounds((int) currentX, (int) currentY, rightX - (int) currentX, height);
				currentX += widgetWidth;
			} else {
				float widgetHeight = widget.getPreferredHeight() * yRatio;
				int bottomY = (int) (currentY + widgetHeight);
				widget.setBounds((int) currentX, (int) currentY, width, bottomY - (int) currentY);
				currentY += widgetHeight;
			}
		}
	}

	@Override
	public void updateViewport(int x, int y, int width, int height) {
		removeNoLongerVisibleItems(x, y, width, height);
		addNewlyVisibleItems(x, y, width, height);
		// System.out.println("*F " + this.lastFirstVisible);
		// contained();
	}

	/**
	 * Adds newly visible items.
	 *
	 * @param x
	 *            the new viewport x coordinate.
	 * @param y
	 *            the new viewport y coordinate.
	 * @param width
	 *            the new viewport width.
	 * @param height
	 *            the new viewport height.
	 */
	private void addNewlyVisibleItems(int x, int y, int width, int height) {
		// Add newly visible items.
		java.util.List<Widget> childrenList = this.childrenList;
		int size = childrenList.size();
		if (size > 0) {
			// System.out.print("ListComposite.updateViewport()");
			// System.out.print(" x=" + x);
			// System.out.print(", y=" + y);
			// System.out.print(", width=" + width);
			// System.out.println(", height=" + height);
			// System.out.println("lastFirstVisible=" + lastFirstVisible);
			int firstVisible = getFirstVisible(x, y, width, height);
			// System.out.println("firstVisible=" + firstVisible);
			boolean horizontal = this.horizontal;
			for (int i = firstVisible - 1; ++i < size;) {
				Widget child = childrenList.get(i);
				int childX = child.getX();
				int childY = child.getY();
				if (horizontal ? (childX + x > width) : (childY + y > height)) {
					// The widget is not in the window, stop recursion.
					break;
				} else {
					int childWidth = child.getWidth();
					int childHeight = child.getHeight();
					if (horizontal ? (childX + childWidth > -x) : (childY + childHeight > -y)) {
						// The widget is in the window.
						if (child.getParent() != this) {
							super.add(child);
							// Force revalidation once the widget is in the hierarchy (Style may be sensible to it).
							child.validate(childWidth, childHeight);
						}
					}
				}
			}
		}
	}

	/**
	 * Removes no more visible items.
	 *
	 * @param x
	 *            the new viewport x coordinate.
	 * @param y
	 *            the new viewport y coordinate.
	 * @param width
	 *            the new viewport width.
	 * @param height
	 *            the new viewport height.
	 */
	private void removeNoLongerVisibleItems(int x, int y, int width, int height) {
		java.util.List<Widget> childrenList = this.childrenList;
		int size = childrenList.size();
		if (size > 0) {
			int firstVisible = this.lastFirstVisible;
			int lastVisible = firstVisible + super.getWidgetsCount();
			boolean horizontal = this.horizontal;
			for (int i = firstVisible - 1; ++i < lastVisible;) {
				Widget child = childrenList.get(i);
				if (horizontal) {
					int childX = child.getX();
					if (childX + x > width || childX + child.getWidth() < -x) {
						super.remove(child);
					}
				} else {
					int childY = child.getY();
					if (childY + y > height || childY + child.getHeight() < -y) {
						super.remove(child);
					}
				}
			}
		}
	}

	// private boolean contains(Widget w) {
	// Iterator<Widget> iterator = super.iterator();
	// while (iterator.hasNext()) {
	// Widget widget = iterator.next();
	// if (widget == w) {
	// return true;
	// }
	// }
	// return false;
	// }
	//
	// private void contained() {
	// Iterator<Widget> iterator = super.iterator();
	// while (iterator.hasNext()) {
	// Widget widget = iterator.next();
	// System.out.println(widget);
	// }
	// }

	private int getFirstVisible(int x, int y, int width, int height) {
		int index = this.lastFirstVisible;
		Widget lastFirstVisibleChild = this.childrenList.get(index);
		if (this.horizontal) {
			getFirstVisibleHorizontally(x, lastFirstVisibleChild);
		} else {
			getFirstVisibleVertically(y, lastFirstVisibleChild);
		}

		return this.lastFirstVisible;
	}

	private void getFirstVisibleVertically(int y, Widget lastFirstVisibleChild) {
		// Be aware that the code is quite the same as getFirstVisibleHorizontally method.

		int childY = lastFirstVisibleChild.getY();
		int childHeight = lastFirstVisibleChild.getHeight();
		// System.out.println("childY=" + childY);
		// System.out.println("childHeight=" + childHeight);
		boolean stillFirst = childY <= -y && childY + childHeight >= -y;
		// System.out.println("childY <= -y=" + (childY <= -y));
		// System.out.println("childY + childHeight >= -y=" + (childY + childHeight >= -y));
		// System.out.println("stillFirst=" + stillFirst);
		if (!stillFirst) {
			boolean searchForward = childY + childHeight < -y;
			// System.out.println("searchForward=" + searchForward);
			java.util.List<Widget> childrenList = this.childrenList;
			int size = childrenList.size();
			int firstCandidate = searchForward ? size - 1 : 0;
			for (int i = this.lastFirstVisible; searchForward ? ++i < size : --i >= 0;) {
				Widget child = childrenList.get(i);
				int candidateY = child.getY();
				int candidateHeight = child.getHeight();
				// System.out.println("i=" + i);
				// System.out.println("y=" + y);
				// System.out.println("candidateY=" + candidateY);
				// System.out.println("candidateHeight=" + candidateHeight);
				if (candidateY <= -y && candidateY + candidateHeight >= -y) {
					firstCandidate = i;
					break;
				}
			}
			this.lastFirstVisible = firstCandidate;
		}
	}

	private void getFirstVisibleHorizontally(int x, Widget lastFirstVisibleChild) {
		// Be aware that the code is quite the same as getFirstVisibleVertically method.

		int childX = lastFirstVisibleChild.getX();
		int childWidth = lastFirstVisibleChild.getWidth();
		boolean stillFirst = childX <= -x && childX + childWidth >= -x;
		if (!stillFirst) {
			boolean searchForward = childX + childWidth < -x;
			java.util.List<Widget> childrenList = this.childrenList;
			int size = childrenList.size();
			int firstCandidate = searchForward ? size - 1 : 0;
			for (int i = this.lastFirstVisible; searchForward ? ++i < size : --i >= 0;) {
				Widget child = childrenList.get(i);
				int candidateX = child.getX();
				int candidateWidth = child.getWidth();
				if (candidateX <= -x && candidateX + candidateWidth >= -x) {
					firstCandidate = i;
					break;
				}
			}
			this.lastFirstVisible = firstCandidate;
		}
	}

}
