/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.navigation.navigator;

import ej.widget.navigation.Navigator;
import ej.widget.navigation.page.ClassNameURLResolver;
import ej.widget.navigation.page.Page;
import ej.widget.navigation.page.URLResolver;
import ej.widget.navigation.stack.PageStack;
import ej.widget.navigation.stack.PageStackReference;

/**
 * Manages the navigation between several pages, from one page to any of the previously visited ones or to a new one.
 * 
 * @deprecated see {@link ej.widget.container.transition.TransitionContainer}.
 */
@Deprecated
public class HistorizedNavigator extends Navigator {

	/**
	 * The stack of pages.
	 */
	protected final PageStack pageStack;
	private final URLResolver urlResolver;

	/**
	 * Creates a historized navigator with a class name URL resolver and a pages stack with references.
	 *
	 * @see ClassNameURLResolver
	 * @see PageStackReference
	 */
	public HistorizedNavigator() {
		this.urlResolver = new ClassNameURLResolver();
		this.pageStack = new PageStackReference(this.urlResolver);
	}

	/**
	 * Creates a historized navigator.
	 *
	 * @param urlResolver
	 *            the URL resolver.
	 * @param pagesStack
	 *            the pages stack.
	 */
	public HistorizedNavigator(URLResolver urlResolver, PageStack pagesStack) {
		this.urlResolver = urlResolver;
		this.pageStack = pagesStack;
	}

	/**
	 * Shows a new page. If there is already a page shown, it is hidden.
	 *
	 * @param url
	 *            the URL of the page to show.
	 */
	public void show(String url) {
		Page page = this.urlResolver.resolve(url);
		show(page);
	}

	/**
	 * Shows a new page. If there is already a page shown, it is hidden.
	 *
	 * @param page
	 *            the page to show.
	 */
	public void show(Page page) {
		Page currentPage = getCurrentPage();
		if (currentPage != null) {
			this.pageStack.push(currentPage);
		}
		show(page, true);
	}

	/**
	 * Reloads the current page.
	 * <p>
	 * A new page with the same URL is created and replaces the current one.
	 */
	public void reload() {
		String url = getCurrentPage().getCurrentURL();
		Page page = this.urlResolver.resolve(url);
		// Copy show() method to avoid possible animation.
		updateCurrentPage(page);
		setChildBounds(page);
		repaint();
	}

	/**
	 * Checks whether or not there are pages in the history.
	 * <p>
	 * The navigation history can be empty if the current page is the first one shown or if the history has just been
	 * cleaned.
	 *
	 * @return <code>true</code> if the navigation history is not empty, <code>false</code> otherwise.
	 * @see #clearHistory()
	 */
	@Override
	public boolean canGoBackward() {
		return this.pageStack.size() > 0;
	}

	/**
	 * Shows the previous page.
	 * <p>
	 * If the navigation history is empty, nothing is done.
	 *
	 * @see #canGoBackward()
	 */
	public void back() {
		if (canGoBackward()) {
			Page previousPage = this.pageStack.pop();
			show(previousPage, false);
		}
	}

	/**
	 * Shows the last page shown with a URL.
	 *
	 * @param url
	 *            the URL to search.
	 *
	 * @see PageStack#popUntil(String)
	 */
	public void backUntil(String url) {
		Page previousPage = this.pageStack.popUntil(url);
		show(previousPage, false);
	}

	/**
	 * Gets the navigation history size.
	 *
	 * @return the navigation history size.
	 * @see PageStack#size()
	 */
	public int getHistorySize() {
		return this.pageStack.size();
	}

	/**
	 * Clears navigation history. The current page remains shown.
	 *
	 * @see PageStack#removeAll()
	 */
	public void clearHistory() {
		this.pageStack.removeAll();
	}

	@Override
	protected Page getPreviousPage() {
		return this.pageStack.peek();
	}

	@Override
	protected void goBackward() {
		back();
	}

}
