/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package googlecloud.iotcore.pubsub;

import java.util.Properties;

import org.eclipse.paho.client.mqttv3.IMqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.MqttCallback;
import org.eclipse.paho.client.mqttv3.MqttMessage;

/**
 * A specific Google Cloud IoT Core client.
 */
public class Client {

	private static final String CLIENT_NOT_CONNECTED = "Client not connected.";
	private static final String CLIENT_ALREADY_CONNECTED = "Client already connected.";
	private static final String NO_TOPIC_LISTENER = "No topic listener installed.";

	// the underlying communication is based on MQTT Paho
	private org.eclipse.paho.client.mqttv3.MqttClient client;

	private TopicListener listener;

	private char[] jwt;

	/**
	 *
	 * @return the state of connection true if connected false otherwise
	 */
	public boolean isConnected() {

		if (this.client == null) {
			return false;
		}

		return this.client.isConnected();
	}

	/**
	 * Use the connection options to open a connection the Google Cloud IoT Core services.
	 *
	 * @param connectOptions
	 *            the specific Google Cloud IoT Core options
	 * @throws IotCoreException
	 *             when already connected and on communication errors
	 */
	public void connect(ConnectOptions connectOptions) throws IotCoreException {

		if (isConnected()) {
			throw new IotCoreException(CLIENT_ALREADY_CONNECTED);
		}

		try {

			org.eclipse.paho.client.mqttv3.MqttConnectOptions options = new org.eclipse.paho.client.mqttv3.MqttConnectOptions();

			options.setMqttVersion(org.eclipse.paho.client.mqttv3.MqttConnectOptions.MQTT_VERSION_3_1_1);

			Properties sslProps = new Properties();
			sslProps.setProperty("com.ibm.ssl.protocol", "TLSv1.2");
			options.setSSLProperties(sslProps);
			options.setSocketFactory(connectOptions.getSslContext().getSocketFactory());

			options.setKeepAliveInterval(0);

			options.setConnectionTimeout(0);

			options.setUserName("unused");

			if (this.jwt == null) {
				char[] jwt = connectOptions.getJWT();

				this.jwt = jwt;
			}

			options.setPassword(this.jwt);

			this.client = new org.eclipse.paho.client.mqttv3.MqttClient(connectOptions.buildServerAddress(),
					connectOptions.buildMqttClientId());

			MqttCallback callback = new MqttCallback() {

				@Override
				public void messageArrived(String topic, MqttMessage message) throws Exception {

					if (Client.this.listener == null) {
						// maybe we never want to get messages from the broker, publish only mode
						// so we never fails here
						return;
					}

					Client.this.listener.onDataReceived(topic, message.getPayload());
				}

				@Override
				public void deliveryComplete(IMqttDeliveryToken token) {
					// only supporting QoS0
					// in QoS 0 this method is only called when written in the underlying socket
					// should never fails
					// so we do nothing here
				}

				@Override
				public void connectionLost(Throwable cause) {
					// the pub/sub methods check the connection state
					// just inform of the lost connection
					cause.printStackTrace();

				}
			};

			this.client.setCallback(callback);

			this.client.connect(options);

		} catch (Throwable e) {
			throw new IotCoreException(e);
		}

	}

	/**
	 * Gets the JSON Web Token once connected.
	 *
	 * @return the jwt.
	 */
	public char[] getJwt() {
		return this.jwt;
	}

	/**
	 * Sets the JSON Web Token to be used for connection.
	 *
	 * @param jwt
	 *            the jwt to used for connection.
	 */
	public void setJwt(char[] jwt) {
		this.jwt = jwt;
	}

	/**
	 * Disconnect this client from the underlying communication layer.
	 *
	 * @throws IotCoreException
	 *             when not connected
	 */
	public void disconnect() throws IotCoreException {

		if (!isConnected()) {
			throw new IotCoreException(CLIENT_NOT_CONNECTED);
		}

		try {
			this.client.disconnect();
		} catch (Throwable t) {
			throw new IotCoreException(t);
		}

	}

	/**
	 *
	 * @param topic
	 *            the topic you want to listened to
	 * @throws IotCoreException
	 *             when not connected and when no TopicListener was installed
	 */
	public void subscribe(String topic) throws IotCoreException {

		if (!isConnected()) {
			throw new IotCoreException(CLIENT_NOT_CONNECTED);
		}

		if (this.listener == null) {
			throw new IotCoreException(NO_TOPIC_LISTENER);
		}

		try {
			this.client.subscribe(topic, 0);
		} catch (Throwable t) {
			throw new IotCoreException(t);
		}
	}

	/**
	 *
	 * @param topic
	 *            the topic you want to stop listening to
	 * @throws IotCoreException
	 *             when not connected
	 */
	public void unsubscribe(String topic) throws IotCoreException {
		if (!isConnected()) {
			throw new IotCoreException(CLIENT_NOT_CONNECTED);
		}

		try {
			this.client.unsubscribe(topic);
		} catch (Throwable t) {
			throw new IotCoreException(t);
		}
	}

	/**
	 *
	 * @param topic
	 *            the topic you want to publish on
	 * @param data
	 *            the data you want to publish
	 * @throws IotCoreException
	 *             when not connected
	 */
	public void publish(String topic, byte[] data) throws IotCoreException {

		if (!isConnected()) {
			throw new IotCoreException(CLIENT_NOT_CONNECTED);
		}

		try {
			this.client.publish(topic, data, 0, false);
		} catch (Throwable t) {
			throw new IotCoreException(t);
		}
	}

	/**
	 *
	 * @param listener
	 *            the TopicListener that will process the received data
	 */
	public void setTopicListener(TopicListener listener) {
		this.listener = listener;
	}

}
